import collections.abc
import inspect
import typing
import unittest

from gi.repository import GIMarshallingTests, GLib, Gio, GObject


class Test(unittest.TestCase):
    def test_signature_attr(self):
        func = GIMarshallingTests.Object.full_inout
        self.assertTrue(hasattr(func, "__signature__"))
        sig1 = func.__signature__
        sig2 = inspect.signature(func)
        self.assertIsInstance(sig1, inspect.Signature)
        self.assertEqual(sig1, sig2)

    def assertSignatureEqual(self, first, second, msg=None):
        first_sig = inspect.signature(first)
        second_sig = inspect.signature(second)
        self.assertEqual(first_sig, second_sig, msg=msg)

    def test_allow_none_with_user_data_defaults(self):
        def expected(
            self,
            destination: Gio.File,
            flags: Gio.FileCopyFlags,
            cancellable: Gio.Cancellable | None = None,
            progress_callback: collections.abc.Callable[[int, int, typing.Any], None]
            | None = None,
            progress_callback_data: typing.Any = None,
        ) -> bool:
            return True

        self.assertSignatureEqual(Gio.File.copy, expected)

    def test_init_function(self):
        def expected(
            argv: list[str] | None = None,
        ) -> tuple[bool, list[str] | None]:
            return (True, [])

        self.assertSignatureEqual(GIMarshallingTests.init_function, expected)

    def test_in_arg(self):
        def expected(ints: list[int]) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.array_in, expected)

    def test_inout_arg(self):
        def expected(ints: list[int]) -> list[int]:
            return []

        self.assertSignatureEqual(GIMarshallingTests.array_inout, expected)

    def test_out_arg(self):
        def expected() -> list[int]:
            return []

        self.assertSignatureEqual(GIMarshallingTests.array_out, expected)

    def test_arg_boolean(self):
        def expected(v: bool) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.boolean_in_true, expected)

    def test_arg_int8(self):
        def expected(v: int) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.int8_in_max, expected)

    def test_arg_uint8(self):
        def expected(v: int) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.uint8_in, expected)

    def test_arg_int16(self):
        def expected(v: int) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.int16_in_max, expected)

    def test_arg_uint16(self):
        def expected(v: int) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.uint16_in, expected)

    def test_arg_int32(self):
        def expected(v: int) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.int32_in_max, expected)

    def test_arg_uint32(self):
        def expected(v: int) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.uint32_in, expected)

    def test_arg_int64(self):
        def expected(v: int) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.int64_in_max, expected)

    def test_arg_uint64(self):
        def expected(v: int) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.uint64_in, expected)

    def test_arg_float(self):
        def expected(v: float) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.float_in, expected)

    def test_arg_double(self):
        def expected(v: float) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.double_in, expected)

    def test_arg_gtype(self):
        def expected(gtype: GObject.GType) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.gtype_in, expected)

    def test_arg_utf8(self):
        def expected(utf8: str) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.utf8_none_in, expected)

    def test_arg_filename(self):
        def expected(path: str) -> bool:
            return True

        self.assertSignatureEqual(GIMarshallingTests.filename_exists, expected)

    def test_arg_enum(self):
        def expected(v: GIMarshallingTests.Enum) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.enum_in, expected)

    def test_arg_genum(self):
        def expected(v: GIMarshallingTests.GEnum) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.genum_in, expected)

    def test_arg_flags(self):
        def expected(v: GIMarshallingTests.NoTypeFlags) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.no_type_flags_in, expected)

    def test_arg_gflags(self):
        def expected(v: GIMarshallingTests.Flags) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.flags_in, expected)

    def test_arg_array(self):
        def expected(ints: list[int]) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.array_fixed_int_in, expected)
        self.assertSignatureEqual(GIMarshallingTests.array_in, expected)

    def test_arg_garray(self):
        def expected(array_: list[int]) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.garray_int_none_in, expected)

    def test_arg_ptrarray(self):
        def expected(parray_: list[str]) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.gptrarray_utf8_none_in, expected)

    def test_arg_glist(self):
        def expected(list: list[int]) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.glist_int_none_in, expected)

    def test_arg_gslist(self):
        def expected(list: list[int]) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.gslist_int_none_in, expected)

    def test_arg_ghashtable(self):
        def expected(hash_table: dict[int, int]) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.ghashtable_int_none_in, expected)

    def test_arg_gvalue(self):
        def expected(value: GObject.Value) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.gvalue_in, expected)

    def test_arg_struct(self):
        def expected(structs: list[GIMarshallingTests.SimpleStruct]) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.array_simple_struct_in, expected)

    def test_arg_boxed(self):
        def expected(structs: list[GIMarshallingTests.BoxedStruct]) -> None:
            pass

        self.assertSignatureEqual(GIMarshallingTests.array_struct_in, expected)

    def test_return(self):
        def expected() -> int:
            return 0

        self.assertSignatureEqual(GIMarshallingTests.int_return_max, expected)

    def test_return_multiple(self):
        def expected() -> tuple[int, int]:
            return (0, 0)

        self.assertSignatureEqual(GIMarshallingTests.int_out_out, expected)
        self.assertSignatureEqual(GIMarshallingTests.int_return_out, expected)

    def test_object_full_inout(self):
        class A:
            @staticmethod
            def expected(
                object: GIMarshallingTests.Object,
            ) -> GIMarshallingTests.Object:
                pass

        self.assertSignatureEqual(GIMarshallingTests.Object.full_inout, A.expected)
        self.assertSignatureEqual(GIMarshallingTests.Object().full_inout, A().expected)

    def test_object_constructor(self):
        class A:
            @classmethod
            def expected(cls, int_: int) -> GIMarshallingTests.Object:
                pass

        self.assertSignatureEqual(GIMarshallingTests.Object.new, A.expected)
        self.assertSignatureEqual(GIMarshallingTests.Object().new, A().expected)

    def test_object_method(self):
        class A:
            def expected(self, ints: list[int]) -> None:
                pass

        self.assertSignatureEqual(GIMarshallingTests.Object.method_array_in, A.expected)
        self.assertSignatureEqual(
            GIMarshallingTests.Object().method_array_in, A().expected
        )

    def test_object_static_method(self):
        class A:
            @staticmethod
            def expected() -> None:
                pass

        self.assertSignatureEqual(GIMarshallingTests.Object.static_method, A.expected)
        self.assertSignatureEqual(
            GIMarshallingTests.Object().static_method, A().expected
        )

    def test_object_virtual_method(self):
        class A:
            @classmethod
            def expected(cls, self, object: GObject.GObject) -> None:
                pass

        self.assertSignatureEqual(
            GIMarshallingTests.Object.do_vfunc_in_object_transfer_none, A.expected
        )
        self.assertSignatureEqual(
            GIMarshallingTests.Object().do_vfunc_in_object_transfer_none, A().expected
        )

    def test_arg_conflict(self):
        class A:
            @classmethod
            def expected(
                type_,
                self,
                offset: int,
                type: GLib.SeekType,
                cancellable: Gio.Cancellable | None = None,
            ) -> bool:
                return True

        self.assertSignatureEqual(Gio.FileIOStream.do_seek, A.expected)



import unittest

import gi.docstring

from gi.repository import Regress
from gi.repository import GIMarshallingTests
from gi.repository import Gio
from gi.repository import GObject

try:
    from gi.repository import Gtk
except ImportError:
    Gtk = None


class Test(unittest.TestCase):
    def test_api(self):
        def new_func(info):
            return "docstring test"

        old_func = gi.docstring.get_doc_string_generator()

        gi.docstring.set_doc_string_generator(new_func)
        self.assertEqual(gi.docstring.get_doc_string_generator(), new_func)
        self.assertEqual(gi.docstring.generate_doc_string(None), "docstring test")

        # Set back to original generator
        gi.docstring.set_doc_string_generator(old_func)
        self.assertEqual(gi.docstring.get_doc_string_generator(), old_func)

    def test_final_signature_with_full_inout(self):
        self.assertEqual(
            GIMarshallingTests.Object.full_inout.__doc__,
            "full_inout(object:GIMarshallingTests.Object) -> object:GIMarshallingTests.Object",
        )

    def test_overridden_doc_is_not_clobbered(self):
        self.assertEqual(
            GIMarshallingTests.OverridesObject.method.__doc__, "Overridden doc string."
        )

    def test_allow_none_with_user_data_defaults(self):
        g_file_copy_doc = (
            "copy(self, destination:Gio.File, "
            "flags:Gio.FileCopyFlags, "
            "cancellable:Gio.Cancellable=None, "
            "progress_callback:Gio.FileProgressCallback=None, "
            "progress_callback_data=None) -> bool"
        )

        self.assertEqual(Gio.File.copy.__doc__, g_file_copy_doc)

    def test_array_length_arg(self):
        self.assertEqual(GIMarshallingTests.array_in.__doc__, "array_in(ints:list)")

    def test_init_function(self):
        # This tests implicit array length args along with skipping a
        # boolean return
        self.assertEqual(
            GIMarshallingTests.init_function.__doc__,
            "init_function(argv:list=None) -> bool, argv:list",
        )

    def test_boolean_return(self):
        self.assertEqual(
            GIMarshallingTests.boolean_return_true.__doc__,
            "boolean_return_true() -> bool",
        )

    # https://bugzilla.gnome.org/show_bug.cgi?id=740301
    def test_may_return_none(self):
        self.assertEqual(
            Gio.File.get_basename.__doc__, "get_basename(self) -> str or None"
        )

    def test_class_doc_constructors(self):
        doc = GIMarshallingTests.Object.__doc__
        self.assertTrue("new(int_:int)" in doc)

    def test_struct_doc_constructors(self):
        doc = GIMarshallingTests.BoxedStruct.__doc__
        self.assertTrue("new()" in doc)
        self.assertTrue("BoxedStruct()" in doc)

    def test_private_struct_constructors(self):
        # Structs without a size or constructor should have no constructor docs.
        doc = Regress.TestBoxedPrivate.__doc__
        self.assertEqual(doc, "")

    def test_array_inout_etc(self):
        self.assertEqual(
            GIMarshallingTests.array_inout_etc.__doc__,
            "array_inout_etc(first:int, ints:list, last:int) -> ints:list, sum:int",
        )

    def test_array_out_etc(self):
        self.assertEqual(
            GIMarshallingTests.array_out_etc.__doc__,
            "array_out_etc(first:int, last:int) -> ints:list, sum:int",
        )

    @unittest.skipUnless(Gtk, "no Gtk")
    def test_shared_array_length_with_prior_out_arg(self):
        # Test the 'iter' out argument does not effect length argument skipping.
        self.assertRegex(
            Gtk.ListStore.insert_with_valuesv.__doc__,
            "insert_with_values.*\\(self, position:int, columns:list, values:list\\) -> iter:Gtk.TreeIter",
        )

    def test_sub_class_doc(self):
        class A(GObject.Object):
            """first doc."""

            pass

        class B(A):
            """second doc."""

            pass

        self.assertEqual(A.__doc__, "first doc.")
        self.assertEqual(B.__doc__, "second doc.")

    def test_sub_class_no_doc(self):
        class A(GObject.Object):
            pass

        class B(A):
            """sub-class doc."""

        self.assertEqual(A.__doc__, None)
        self.assertEqual(B.__doc__, "sub-class doc.")

    @unittest.expectedFailure  # https://bugzilla.gnome.org/show_bug.cgi?id=734926
    def test_sub_class_doc_setattr(self):
        class A(GObject.Object):
            pass

        class B(A):
            pass

        A.__doc__ = "custom doc"

        self.assertEqual(A.__doc__, "custom doc")
        self.assertEqual(B.__doc__, "custom doc")

    def test_return_array_with_length_argument(self):
        self.assertEqual(
            GIMarshallingTests.enum_array_return_type.__doc__,
            "enum_array_return_type() -> list",
        )



import os
import unittest
import tempfile
import os.path
import shutil
import warnings

try:
    import fcntl
except ImportError:
    fcntl = None

from gi.repository import GLib
from gi import PyGIDeprecationWarning


class IOChannel(unittest.TestCase):
    def setUp(self):
        self.workdir = tempfile.mkdtemp()

        self.testutf8 = os.path.join(self.workdir, "testutf8.txt")
        with open(self.testutf8, "wb") as f:
            f.write(
                """hello ♥ world
second line

À demain!""".encode()
            )

        self.testlatin1 = os.path.join(self.workdir, "testlatin1.txt")
        with open(self.testlatin1, "wb") as f:
            f.write(b"""hell\xf8 world
second line

\xc0 demain!""")

        self.testout = os.path.join(self.workdir, "testout.txt")

    def tearDown(self):
        shutil.rmtree(self.workdir)

    def test_file_readline_utf8(self):
        ch = GLib.IOChannel(filename=self.testutf8)
        self.assertEqual(ch.get_encoding(), "UTF-8")
        self.assertTrue(ch.get_close_on_unref())
        self.assertEqual(ch.readline(), "hello ♥ world\n")
        self.assertEqual(ch.get_buffer_condition(), GLib.IOCondition.IN)
        self.assertEqual(ch.readline(), "second line\n")
        self.assertEqual(ch.readline(), "\n")
        self.assertEqual(ch.readline(), "À demain!")
        self.assertEqual(ch.get_buffer_condition(), 0)
        self.assertEqual(ch.readline(), "")
        ch.shutdown(True)

    def test_file_readline_latin1(self):
        ch = GLib.IOChannel(filename=self.testlatin1, mode="r")
        ch.set_encoding("latin1")
        self.assertEqual(ch.get_encoding(), "latin1")
        self.assertEqual(ch.readline(), "hellø world\n")
        self.assertEqual(ch.readline(), "second line\n")
        self.assertEqual(ch.readline(), "\n")
        self.assertEqual(ch.readline(), "À demain!")
        ch.shutdown(True)

    def test_file_iter(self):
        items = []
        ch = GLib.IOChannel(filename=self.testutf8)
        for item in ch:
            items.append(item)  # noqa: PERF402
        self.assertEqual(len(items), 4)
        self.assertEqual(items[0], "hello ♥ world\n")
        ch.shutdown(True)

    def test_file_readlines(self):
        ch = GLib.IOChannel(filename=self.testutf8)
        lines = ch.readlines()
        # Note, this really ought to be 4, but the static bindings add an extra
        # empty one
        self.assertGreaterEqual(len(lines), 4)
        self.assertLessEqual(len(lines), 5)
        self.assertEqual(lines[0], "hello ♥ world\n")
        self.assertEqual(lines[3], "À demain!")
        if len(lines) == 4:
            self.assertEqual(lines[4], "")

    def test_file_read(self):
        ch = GLib.IOChannel(filename=self.testutf8)
        with open(self.testutf8, "rb") as f:
            self.assertEqual(ch.read(), f.read())

        ch = GLib.IOChannel(filename=self.testutf8)
        with open(self.testutf8, "rb") as f:
            self.assertEqual(ch.read(10), f.read(10))

        ch = GLib.IOChannel(filename=self.testutf8)
        with open(self.testutf8, "rb") as f:
            self.assertEqual(ch.read(max_count=15), f.read(15))

    def test_file_read_chars(self):
        ch = GLib.IOChannel(filename=self.testutf8)
        with open(self.testutf8, "rb") as f:
            self.assertEqual(ch.read_chars(), f.read())

    def test_seek(self):
        ch = GLib.IOChannel(filename=self.testutf8)
        ch.seek(2)
        self.assertEqual(ch.read(3), b"llo")

        ch.seek(2, 0)  # SEEK_SET
        self.assertEqual(ch.read(3), b"llo")

        ch.seek(1, 1)  # SEEK_CUR, skip the space
        self.assertEqual(ch.read(3), b"\xe2\x99\xa5")

        ch.seek(2, 2)  # SEEK_END
        # FIXME: does not work currently
        # self.assertEqual(ch.read(2), b'n!')

        # invalid whence value
        self.assertRaises(ValueError, ch.seek, 0, 3)
        ch.shutdown(True)

    def test_file_write(self):
        ch = GLib.IOChannel(filename=self.testout, mode="w")
        ch.set_encoding("latin1")
        ch.write("hellø world\n")
        ch.shutdown(True)
        ch = GLib.IOChannel(filename=self.testout, mode="a")
        ch.set_encoding("latin1")
        ch.write("À demain!")
        ch.shutdown(True)

        with open(self.testout, "rb") as f:
            self.assertEqual(f.read().decode("latin1"), "hellø world\nÀ demain!")

    def test_file_writelines(self):
        ch = GLib.IOChannel(filename=self.testout, mode="w")
        ch.writelines(["foo", "bar\n", "baz\n", "end"])
        ch.shutdown(True)

        with open(self.testout) as f:
            self.assertEqual(f.read(), "foobar\nbaz\nend")

    def test_buffering(self):
        writer = GLib.IOChannel(filename=self.testout, mode="w")
        writer.set_encoding(None)
        self.assertTrue(writer.get_buffered())
        self.assertGreater(writer.get_buffer_size(), 10)

        reader = GLib.IOChannel(filename=self.testout, mode="r")

        # does not get written immediately on buffering
        writer.write("abc")
        self.assertEqual(reader.read(), b"")
        writer.flush()
        self.assertEqual(reader.read(), b"abc")

        # does get written immediately without buffering
        writer.set_buffered(False)
        writer.write("def")
        self.assertEqual(reader.read(), b"def")

        # writes after buffer overflow
        writer.set_buffer_size(10)
        writer.write("0123456789012")
        self.assertTrue(reader.read().startswith(b"012"))
        writer.flush()
        reader.read()  # ignore bits written after flushing

        # closing flushes
        writer.set_buffered(True)
        writer.write("ghi")
        writer.shutdown(True)
        self.assertEqual(reader.read(), b"ghi")
        reader.shutdown(True)

    @unittest.skipIf(os.name == "nt", "NONBLOCK not implemented on Windows")
    def test_fd_read(self):
        (r, w) = os.pipe()

        ch = GLib.IOChannel(filedes=r)
        ch.set_encoding(None)
        ch.set_flags(ch.get_flags() | GLib.IOFlags.NONBLOCK)
        self.assertNotEqual(ch.get_flags() | GLib.IOFlags.NONBLOCK, 0)
        self.assertEqual(ch.read(), b"")
        os.write(w, b"\x01\x02")
        self.assertEqual(ch.read(), b"\x01\x02")

        # now test blocking case, after closing the write end
        ch.set_flags(GLib.IOFlags(ch.get_flags() & ~GLib.IOFlags.NONBLOCK))
        os.write(w, b"\x03\x04")
        os.close(w)
        self.assertEqual(ch.read(), b"\x03\x04")

        ch.shutdown(True)

    @unittest.skipUnless(fcntl, "no fcntl")
    def test_fd_write(self):
        (r, w) = os.pipe()
        fcntl.fcntl(r, fcntl.F_SETFL, fcntl.fcntl(r, fcntl.F_GETFL) | os.O_NONBLOCK)

        ch = GLib.IOChannel(filedes=w, mode="w")
        ch.set_encoding(None)
        ch.set_buffered(False)
        ch.write(b"\x01\x02")
        self.assertEqual(os.read(r, 10), b"\x01\x02")

        # now test blocking case, after closing the write end
        fcntl.fcntl(r, fcntl.F_SETFL, fcntl.fcntl(r, fcntl.F_GETFL) & ~os.O_NONBLOCK)
        ch.write(b"\x03\x04")
        ch.shutdown(True)
        self.assertEqual(os.read(r, 10), b"\x03\x04")
        os.close(r)

    @unittest.skipIf(os.name == "nt", "NONBLOCK not implemented on Windows")
    def test_deprecated_method_add_watch_no_data(self):
        (r, w) = os.pipe()

        ch = GLib.IOChannel(filedes=r)
        ch.set_encoding(None)
        ch.set_flags(ch.get_flags() | GLib.IOFlags.NONBLOCK)

        cb_reads = []

        def cb(channel, condition):
            self.assertEqual(channel, ch)
            self.assertEqual(condition, GLib.IOCondition.IN)
            cb_reads.append(channel.read())
            if len(cb_reads) == 2:
                ml.quit()
            return True

        # io_add_watch() method is deprecated, use GLib.io_add_watch
        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            ch.add_watch(GLib.IOCondition.IN, cb, priority=GLib.PRIORITY_HIGH)
            self.assertTrue(issubclass(warn[0].category, PyGIDeprecationWarning))

        def write():
            os.write(w, b"a")
            GLib.idle_add(lambda: os.write(w, b"b") and False)

        ml = GLib.MainLoop()
        GLib.idle_add(write)
        GLib.timeout_add(2000, ml.quit)
        ml.run()

        self.assertEqual(cb_reads, [b"a", b"b"])

    @unittest.skipIf(os.name == "nt", "NONBLOCK not implemented on Windows")
    def test_deprecated_method_add_watch_data_priority(self):
        (r, w) = os.pipe()

        ch = GLib.IOChannel(filedes=r)
        ch.set_encoding(None)
        ch.set_flags(ch.get_flags() | GLib.IOFlags.NONBLOCK)

        cb_reads = []

        def cb(channel, condition, data):
            self.assertEqual(channel, ch)
            self.assertEqual(condition, GLib.IOCondition.IN)
            self.assertEqual(data, "hello")
            cb_reads.append(channel.read())
            if len(cb_reads) == 2:
                ml.quit()
            return True

        ml = GLib.MainLoop()
        # io_add_watch() method is deprecated, use GLib.io_add_watch
        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            id = ch.add_watch(
                GLib.IOCondition.IN, cb, "hello", priority=GLib.PRIORITY_HIGH
            )
            self.assertTrue(issubclass(warn[0].category, PyGIDeprecationWarning))

        self.assertEqual(
            ml.get_context().find_source_by_id(id).priority, GLib.PRIORITY_HIGH
        )

        def write():
            os.write(w, b"a")
            GLib.idle_add(lambda: os.write(w, b"b") and False)

        GLib.idle_add(write)
        GLib.timeout_add(2000, ml.quit)
        ml.run()

        self.assertEqual(cb_reads, [b"a", b"b"])

    @unittest.skipIf(os.name == "nt", "NONBLOCK not implemented on Windows")
    def test_add_watch_no_data(self):
        (r, w) = os.pipe()

        ch = GLib.IOChannel(filedes=r)
        ch.set_encoding(None)
        ch.set_flags(ch.get_flags() | GLib.IOFlags.NONBLOCK)

        cb_reads = []

        def cb(channel, condition):
            self.assertEqual(channel, ch)
            self.assertEqual(condition, GLib.IOCondition.IN)
            cb_reads.append(channel.read())
            if len(cb_reads) == 2:
                ml.quit()
            return True

        id = GLib.io_add_watch(ch, GLib.PRIORITY_HIGH, GLib.IOCondition.IN, cb)

        ml = GLib.MainLoop()
        self.assertEqual(
            ml.get_context().find_source_by_id(id).priority, GLib.PRIORITY_HIGH
        )

        def write():
            os.write(w, b"a")
            GLib.idle_add(lambda: os.write(w, b"b") and False)

        GLib.idle_add(write)
        GLib.timeout_add(2000, ml.quit)
        ml.run()

        self.assertEqual(cb_reads, [b"a", b"b"])

    @unittest.skipIf(os.name == "nt", "NONBLOCK not implemented on Windows")
    def test_add_watch_with_data(self):
        (r, w) = os.pipe()

        ch = GLib.IOChannel(filedes=r)
        ch.set_encoding(None)
        ch.set_flags(ch.get_flags() | GLib.IOFlags.NONBLOCK)

        cb_reads = []

        def cb(channel, condition, data):
            self.assertEqual(channel, ch)
            self.assertEqual(condition, GLib.IOCondition.IN)
            self.assertEqual(data, "hello")
            cb_reads.append(channel.read())
            if len(cb_reads) == 2:
                ml.quit()
            return True

        id = GLib.io_add_watch(ch, GLib.PRIORITY_HIGH, GLib.IOCondition.IN, cb, "hello")

        ml = GLib.MainLoop()
        self.assertEqual(
            ml.get_context().find_source_by_id(id).priority, GLib.PRIORITY_HIGH
        )

        def write():
            os.write(w, b"a")
            GLib.idle_add(lambda: os.write(w, b"b") and False)

        GLib.idle_add(write)
        GLib.timeout_add(2000, ml.quit)
        ml.run()

        self.assertEqual(cb_reads, [b"a", b"b"])

    @unittest.skipIf(os.name == "nt", "NONBLOCK not implemented on Windows")
    def test_add_watch_with_multi_data(self):
        (r, w) = os.pipe()

        ch = GLib.IOChannel(filedes=r)
        ch.set_encoding(None)
        ch.set_flags(ch.get_flags() | GLib.IOFlags.NONBLOCK)

        cb_reads = []

        def cb(channel, condition, data1, data2, data3):
            self.assertEqual(channel, ch)
            self.assertEqual(condition, GLib.IOCondition.IN)
            self.assertEqual(data1, "a")
            self.assertEqual(data2, "b")
            self.assertEqual(data3, "c")
            cb_reads.append(channel.read())
            if len(cb_reads) == 2:
                ml.quit()
            return True

        id = GLib.io_add_watch(
            ch, GLib.PRIORITY_HIGH, GLib.IOCondition.IN, cb, "a", "b", "c"
        )

        ml = GLib.MainLoop()
        self.assertEqual(
            ml.get_context().find_source_by_id(id).priority, GLib.PRIORITY_HIGH
        )

        def write():
            os.write(w, b"a")
            GLib.idle_add(lambda: os.write(w, b"b") and False)

        GLib.idle_add(write)
        GLib.timeout_add(2000, ml.quit)
        ml.run()

        self.assertEqual(cb_reads, [b"a", b"b"])

    @unittest.skipIf(os.name == "nt", "NONBLOCK not implemented on Windows")
    def test_deprecated_add_watch_no_data(self):
        (r, w) = os.pipe()

        ch = GLib.IOChannel(filedes=r)
        ch.set_encoding(None)
        ch.set_flags(ch.get_flags() | GLib.IOFlags.NONBLOCK)

        cb_reads = []

        def cb(channel, condition):
            self.assertEqual(channel, ch)
            self.assertEqual(condition, GLib.IOCondition.IN)
            cb_reads.append(channel.read())
            if len(cb_reads) == 2:
                ml.quit()
            return True

        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            id = GLib.io_add_watch(
                ch, GLib.IOCondition.IN, cb, priority=GLib.PRIORITY_HIGH
            )
            self.assertTrue(issubclass(warn[0].category, PyGIDeprecationWarning))

        ml = GLib.MainLoop()
        self.assertEqual(
            ml.get_context().find_source_by_id(id).priority, GLib.PRIORITY_HIGH
        )

        def write():
            os.write(w, b"a")
            GLib.idle_add(lambda: os.write(w, b"b") and False)

        GLib.idle_add(write)
        GLib.timeout_add(2000, ml.quit)
        ml.run()

        self.assertEqual(cb_reads, [b"a", b"b"])

    @unittest.skipIf(os.name == "nt", "NONBLOCK not implemented on Windows")
    def test_deprecated_add_watch_with_data(self):
        (r, w) = os.pipe()

        ch = GLib.IOChannel(filedes=r)
        ch.set_encoding(None)
        ch.set_flags(ch.get_flags() | GLib.IOFlags.NONBLOCK)

        cb_reads = []

        def cb(channel, condition, data):
            self.assertEqual(channel, ch)
            self.assertEqual(condition, GLib.IOCondition.IN)
            self.assertEqual(data, "hello")
            cb_reads.append(channel.read())
            if len(cb_reads) == 2:
                ml.quit()
            return True

        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            id = GLib.io_add_watch(
                ch, GLib.IOCondition.IN, cb, "hello", priority=GLib.PRIORITY_HIGH
            )
            self.assertTrue(issubclass(warn[0].category, PyGIDeprecationWarning))

        ml = GLib.MainLoop()
        self.assertEqual(
            ml.get_context().find_source_by_id(id).priority, GLib.PRIORITY_HIGH
        )

        def write():
            os.write(w, b"a")
            GLib.idle_add(lambda: os.write(w, b"b") and False)

        GLib.idle_add(write)

        GLib.timeout_add(2000, ml.quit)
        ml.run()

        self.assertEqual(cb_reads, [b"a", b"b"])

    def test_backwards_compat_flags(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", PyGIDeprecationWarning)

            self.assertEqual(GLib.IOCondition.IN, GLib.IO_IN)
            self.assertEqual(GLib.IOFlags.NONBLOCK, GLib.IO_FLAG_NONBLOCK)
            self.assertEqual(GLib.IOFlags.IS_SEEKABLE, GLib.IO_FLAG_IS_SEEKABLE)
            self.assertEqual(GLib.IOStatus.NORMAL, GLib.IO_STATUS_NORMAL)



import gc
import unittest
import sys
import weakref
import threading
import time

from gi.repository import GObject, GLib, Regress, Gio
from gi import _signalhelper as signalhelper
from gi.module import repository as repo

import testhelper
from .helper import capture_glib_warnings, capture_gi_deprecation_warnings


class C(GObject.GObject):
    __gsignals__ = {
        "my_signal": (GObject.SignalFlags.RUN_FIRST, None, (GObject.TYPE_INT,))
    }

    def do_my_signal(self, arg):
        self.arg = arg


class D(C):
    def do_my_signal(self, arg2):
        self.arg2 = arg2
        C.do_my_signal(self, arg2)


class TestSignalCreation(unittest.TestCase):
    # Bug 540376.
    def test_illegals(self):
        self.assertRaises(
            TypeError,
            lambda: GObject.signal_new("test", None, 0, None, (GObject.TYPE_LONG,)),
        )


class TestChaining(unittest.TestCase):
    def setUp(self):
        self.inst = C()
        self.inst.connect("my_signal", self.my_signal_handler_cb, 1, 2, 3)

    def my_signal_handler_cb(self, *args):
        assert len(args) == 5
        assert isinstance(args[0], C)
        assert args[0] == self.inst

        assert isinstance(args[1], int)
        assert args[1] == 42

        assert args[2:] == (1, 2, 3)

    def test_chaining(self):
        self.inst.emit("my_signal", 42)
        assert self.inst.arg == 42

    def test_chaining2(self):
        inst2 = D()
        inst2.emit("my_signal", 44)
        assert inst2.arg == 44
        assert inst2.arg2 == 44


# This is for bug 153718


class TestGSignalsError(unittest.TestCase):
    def test_invalid_type(self, *args):
        def foo():
            class Foo(GObject.GObject):
                __gsignals__ = None

        self.assertRaises(TypeError, foo)
        gc.collect()

    def test_invalid_name(self, *args):
        def foo():
            class Foo(GObject.GObject):
                __gsignals__ = {"not-exists": "override"}

        with capture_glib_warnings(allow_warnings=True):
            self.assertRaises(TypeError, foo)
        gc.collect()


class TestGPropertyError(unittest.TestCase):
    def test_invalid_type(self, *args):
        def foo():
            class Foo(GObject.GObject):
                __gproperties__ = None

        self.assertRaises(TypeError, foo)
        gc.collect()

    def test_invalid_name(self, *args):
        def foo():
            class Foo(GObject.GObject):
                __gproperties__ = {None: None}

        self.assertRaises(TypeError, foo)
        gc.collect()


class TestList(unittest.TestCase):
    def test_list_names(self):
        self.assertEqual(GObject.signal_list_names(C), ("my-signal",))


def my_accumulator(ihint, return_accu, handler_return, user_data):
    """An accumulator that stops emission when the sum of handler
    returned values reaches 3.
    """
    assert user_data == "accum data"
    if return_accu >= 3:
        return False, return_accu
    return True, return_accu + handler_return


class Foo(GObject.GObject):
    my_acc_signal = GObject.Signal(
        return_type=GObject.TYPE_INT,
        flags=GObject.SignalFlags.RUN_LAST,
        accumulator=my_accumulator,
        accu_data="accum data",
    )

    my_other_acc_signal = GObject.Signal(
        return_type=GObject.TYPE_BOOLEAN,
        flags=GObject.SignalFlags.RUN_LAST,
        accumulator=GObject.signal_accumulator_true_handled,
    )

    my_acc_first_wins = GObject.Signal(
        return_type=GObject.TYPE_BOOLEAN,
        flags=GObject.SignalFlags.RUN_LAST,
        accumulator=GObject.signal_accumulator_first_wins,
    )


class TestAccumulator(unittest.TestCase):
    def test_accumulator(self):
        inst = Foo()
        inst.my_acc_signal.connect(lambda obj: 1)
        inst.my_acc_signal.connect(lambda obj: 2)
        # the value returned in the following handler will not be
        # considered, because at this point the accumulator already
        # reached its limit.
        inst.my_acc_signal.connect(lambda obj: 3)
        retval = inst.my_acc_signal.emit()
        self.assertEqual(retval, 3)

    def test_accumulator_true_handled(self):
        inst = Foo()
        inst.my_other_acc_signal.connect(self._true_handler1)
        inst.my_other_acc_signal.connect(self._true_handler2)
        # the following handler will not be called because handler2
        # returns True, so it should stop the emission.
        inst.my_other_acc_signal.connect(self._true_handler3)
        self.__true_val = None
        inst.my_other_acc_signal.emit()
        self.assertEqual(self.__true_val, 2)

    def test_accumulator_first_wins(self):
        # First signal hit will always win
        inst = Foo()
        inst.my_acc_first_wins.connect(self._true_handler3)
        inst.my_acc_first_wins.connect(self._true_handler1)
        inst.my_acc_first_wins.connect(self._true_handler2)
        self.__true_val = None
        inst.my_acc_first_wins.emit()
        self.assertEqual(self.__true_val, 3)

    def _true_handler1(self, obj):
        self.__true_val = 1
        return False

    def _true_handler2(self, obj):
        self.__true_val = 2
        return True

    def _true_handler3(self, obj):
        self.__true_val = 3
        return False


class E(GObject.GObject):
    __gsignals__ = {"signal": (GObject.SignalFlags.RUN_FIRST, None, ())}

    # Property used to test detailed signal
    prop = GObject.Property(type=int, default=0)

    def __init__(self):
        GObject.GObject.__init__(self)
        self.status = 0

    def do_signal(self):
        assert self.status == 0
        self.status = 1


class F(GObject.GObject):
    __gsignals__ = {"signal": (GObject.SignalFlags.RUN_FIRST, None, ())}

    def __init__(self):
        GObject.GObject.__init__(self)
        self.status = 0

    def do_signal(self):
        self.status += 1


class TestEmissionHook(unittest.TestCase):
    def test_add(self):
        self.hook = True
        e = E()
        e.connect("signal", self._callback)
        GObject.add_emission_hook(E, "signal", self._emission_hook)
        e.emit("signal")
        self.assertEqual(e.status, 3)

    def test_remove(self):
        self.hook = False
        e = E()
        e.connect("signal", self._callback)
        hook_id = GObject.add_emission_hook(E, "signal", self._emission_hook)
        GObject.remove_emission_hook(E, "signal", hook_id)
        e.emit("signal")
        self.assertEqual(e.status, 3)

    def _emission_hook(self, e):
        self.assertEqual(e.status, 1)
        e.status = 2

    def _callback(self, e):
        if self.hook:
            self.assertEqual(e.status, 2)
        else:
            self.assertEqual(e.status, 1)
        e.status = 3

    def test_callback_return_false(self):
        self.hook = False
        obj = F()

        def _emission_hook(obj):
            obj.status += 1
            return False

        GObject.add_emission_hook(obj, "signal", _emission_hook)
        obj.emit("signal")
        obj.emit("signal")
        self.assertEqual(obj.status, 3)

    def test_callback_return_true(self):
        self.hook = False
        obj = F()

        def _emission_hook(obj):
            obj.status += 1
            return True

        hook_id = GObject.add_emission_hook(obj, "signal", _emission_hook)
        obj.emit("signal")
        obj.emit("signal")
        GObject.remove_emission_hook(obj, "signal", hook_id)
        self.assertEqual(obj.status, 4)

    def test_callback_return_true_but_remove(self):
        self.hook = False
        obj = F()

        def _emission_hook(obj):
            obj.status += 1
            return True

        hook_id = GObject.add_emission_hook(obj, "signal", _emission_hook)
        obj.emit("signal")
        GObject.remove_emission_hook(obj, "signal", hook_id)
        obj.emit("signal")
        self.assertEqual(obj.status, 3)


class TestMatching(unittest.TestCase):
    class Object(GObject.Object):
        status = 0
        prop = GObject.Property(type=int, default=0)

        @GObject.Signal()
        def my_signal(self):
            pass

    @unittest.expectedFailure  # https://bugzilla.gnome.org/show_bug.cgi?id=692918
    def test_signal_handler_block_matching(self):
        def dummy(*args):
            """Hack to work around:"""

        def foo(obj):
            obj.status += 1

        obj = self.Object()
        handler_id = GObject.signal_connect_closure(obj, "my-signal", foo, after=False)
        handler_id

        self.assertEqual(obj.status, 0)
        obj.emit("my-signal")
        self.assertEqual(obj.status, 1)

        # Blocking by match criteria disables the foo callback
        signal_id, detail = GObject.signal_parse_name("my-signal", obj, True)
        count = GObject.signal_handlers_block_matched(
            obj,
            GObject.SignalMatchType.ID | GObject.SignalMatchType.CLOSURE,
            signal_id=signal_id,
            detail=detail,
            closure=foo,
            func=dummy,
            data=dummy,
        )
        self.assertEqual(count, 1)
        obj.emit("my-signal")
        self.assertEqual(obj.status, 1)

        # Unblocking by the same match criteria allows callback to work again
        count = GObject.signal_handlers_unblock_matched(
            obj,
            GObject.SignalMatchType.ID | GObject.SignalMatchType.CLOSURE,
            signal_id=signal_id,
            detail=detail,
            closure=foo,
            func=dummy,
            data=dummy,
        )
        self.assertEqual(count, 1)
        obj.emit("my-signal")
        self.assertEqual(obj.status, 2)

        # Disconnecting by match criteria completely removes the handler
        count = GObject.signal_handlers_disconnect_matched(
            obj,
            GObject.SignalMatchType.ID | GObject.SignalMatchType.CLOSURE,
            signal_id=signal_id,
            detail=detail,
            closure=foo,
            func=dummy,
            data=dummy,
        )
        self.assertEqual(count, 1)
        obj.emit("my-signal")
        self.assertEqual(obj.status, 2)

    def test_signal_handler_find(self):
        def foo(obj):
            obj.status += 1

        obj = self.Object()
        handler_id = GObject.signal_connect_closure(obj, "my-signal", foo, after=False)

        signal_id, detail = GObject.signal_parse_name("my-signal", obj, True)
        found_id = GObject.signal_handler_find(
            obj,
            GObject.SignalMatchType.ID,
            signal_id=signal_id,
            detail=detail,
            closure=None,
            func=0,
            data=0,
        )
        self.assertEqual(handler_id, found_id)


class TestClosures(unittest.TestCase):
    def setUp(self):
        self.count = 0
        self.emission_stopped = False
        self.emission_error = False
        self.handler_pending = False

    def _callback_handler_pending(self, e):
        signal_id, detail = GObject.signal_parse_name("signal", e, True)
        self.handler_pending = GObject.signal_has_handler_pending(
            e, signal_id, detail, may_be_blocked=False
        )

    def _callback(self, e):
        self.count += 1

    def _callback_stop_emission(self, obj, prop, stop_it):
        if stop_it:
            obj.stop_emission_by_name("notify::prop")
            self.emission_stopped = True
        else:
            self.count += 1

    def _callback_invalid_stop_emission_name(self, obj, prop):
        with capture_glib_warnings(allow_warnings=True, allow_criticals=True) as warn:
            obj.stop_emission_by_name("notasignal::baddetail")
            self.emission_error = True
            self.assertTrue(warn)

    def test_disconnect_by_func(self):
        e = E()
        e.connect("signal", self._callback)
        e.disconnect_by_func(self._callback)
        e.emit("signal")
        self.assertEqual(self.count, 0)

    def test_disconnect(self):
        e = E()
        handler_id = e.connect("signal", self._callback)
        self.assertTrue(e.handler_is_connected(handler_id))
        e.disconnect(handler_id)
        e.emit("signal")
        self.assertEqual(self.count, 0)
        self.assertFalse(e.handler_is_connected(handler_id))

    def test_stop_emission_by_name(self):
        e = E()

        # Sandwich a callback that stops emission in between a callback that increments
        e.connect("notify::prop", self._callback_stop_emission, False)
        e.connect("notify::prop", self._callback_stop_emission, True)
        e.connect("notify::prop", self._callback_stop_emission, False)

        e.set_property("prop", 1234)
        self.assertEqual(e.get_property("prop"), 1234)
        self.assertEqual(self.count, 1)
        self.assertTrue(self.emission_stopped)

    def test_stop_emission_by_name_error(self):
        e = E()

        e.connect("notify::prop", self._callback_invalid_stop_emission_name)
        with capture_glib_warnings():
            e.set_property("prop", 1234)
        self.assertTrue(self.emission_error)

    def test_handler_block(self):
        e = E()
        e.connect("signal", self._callback)
        e.handler_block_by_func(self._callback)
        e.emit("signal")
        self.assertEqual(self.count, 0)

    def test_handler_unblock(self):
        e = E()
        handler_id = e.connect("signal", self._callback)
        e.handler_block(handler_id)
        e.handler_unblock_by_func(self._callback)
        e.emit("signal")
        self.assertEqual(self.count, 1)

    def test_handler_block_method(self):
        # Filed as #375589
        class A:
            def __init__(self):
                self.a = 0

            def callback(self, o):
                self.a = 1
                o.handler_block_by_func(self.callback)

        inst = A()
        e = E()
        e.connect("signal", inst.callback)
        e.emit("signal")
        self.assertEqual(inst.a, 1)
        gc.collect()

    def test_gstring(self):
        class C(GObject.GObject):
            __gsignals__ = {
                "my_signal": (
                    GObject.SignalFlags.RUN_LAST,
                    GObject.TYPE_GSTRING,
                    (GObject.TYPE_GSTRING,),
                )
            }

            def __init__(self, test):
                GObject.GObject.__init__(self)
                self.test = test

            def do_my_signal(self, data):
                self.data = data
                self.test.assertEqual(len(data), 3)
                return "".join([data[2], data[1], data[0]])

        c = C(self)
        data = c.emit("my_signal", "\01\00\02")
        self.assertEqual(data, "\02\00\01")

    def test_handler_pending(self):
        obj = F()
        obj.connect("signal", self._callback_handler_pending)
        obj.connect("signal", self._callback)

        self.assertEqual(self.count, 0)
        self.assertEqual(self.handler_pending, False)

        obj.emit("signal")
        self.assertEqual(self.count, 1)
        self.assertEqual(self.handler_pending, True)

    def test_signal_handlers_destroy(self):
        obj = F()
        obj.connect("signal", self._callback)
        obj.connect("signal", self._callback)
        obj.connect("signal", self._callback)

        obj.emit("signal")
        self.assertEqual(self.count, 3)

        # count should remain at 3 after all handlers are destroyed
        GObject.signal_handlers_destroy(obj)
        obj.emit("signal")
        self.assertEqual(self.count, 3)


class SigPropClass(GObject.GObject):
    __gsignals__ = {
        "my_signal": (GObject.SignalFlags.RUN_FIRST, None, (GObject.TYPE_INT,))
    }

    __gproperties__ = {
        "foo": (
            str,
            None,
            None,
            "",
            GObject.ParamFlags.WRITABLE | GObject.ParamFlags.CONSTRUCT,
        ),
    }

    signal_emission_failed = False

    def do_my_signal(self, arg):
        self.arg = arg

    def do_set_property(self, pspec, value):
        if pspec.name == "foo":
            self._foo = value
        else:
            raise AttributeError(f"unknown property {pspec.name}")
        try:
            self.emit("my-signal", 1)
        except TypeError:
            self.signal_emission_failed = True


class TestSigProp(unittest.TestCase):
    def test_emit_in_property_setter(self):
        obj = SigPropClass()
        self.assertFalse(obj.signal_emission_failed)


class CM(GObject.GObject):
    __gsignals__ = {
        "test1": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "test2": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "test3": (GObject.SignalFlags.RUN_LAST, int, (GObject.TYPE_DOUBLE,)),
        "test4": (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (
                bool,
                int,
                GObject.TYPE_FLOAT,
                GObject.TYPE_DOUBLE,
                int,
                GObject.TYPE_UINT,
                GObject.TYPE_ULONG,
            ),
        ),
        "test_float": (
            GObject.SignalFlags.RUN_LAST,
            GObject.TYPE_FLOAT,
            (GObject.TYPE_FLOAT,),
        ),
        "test_double": (
            GObject.SignalFlags.RUN_LAST,
            GObject.TYPE_DOUBLE,
            (GObject.TYPE_DOUBLE,),
        ),
        "test_int64": (
            GObject.SignalFlags.RUN_LAST,
            GObject.TYPE_INT64,
            (GObject.TYPE_INT64,),
        ),
        "test_string": (GObject.SignalFlags.RUN_LAST, str, (str,)),
        "test_object": (GObject.SignalFlags.RUN_LAST, object, (object,)),
        "test_paramspec": (GObject.SignalFlags.RUN_LAST, GObject.ParamSpec, ()),
        "test_paramspec_in": (
            GObject.SignalFlags.RUN_LAST,
            GObject.ParamSpec,
            (GObject.ParamSpec,),
        ),
        "test_gvalue": (GObject.SignalFlags.RUN_LAST, GObject.Value, (GObject.Value,)),
        "test_gvalue_ret": (
            GObject.SignalFlags.RUN_LAST,
            GObject.Value,
            (GObject.TYPE_GTYPE,),
        ),
    }

    testprop = GObject.Property(type=int)


class _TestCMarshaller:
    def setUp(self):
        self.obj = CM()
        testhelper.connectcallbacks(self.obj)

    def test_test1(self):
        self.obj.emit("test1")

    def test_test2(self):
        self.obj.emit("test2", "string")

    def test_test3(self):
        rv = self.obj.emit("test3", 42.0)
        self.assertEqual(rv, 20)

    def test_test4(self):
        self.obj.emit("test4", True, 10, 3.14, 1.78, 20, 30, 31)

    def test_float(self):
        rv = self.obj.emit("test-float", 1.234)
        self.assertTrue(rv >= 1.233999 and rv <= 1.2400001, rv)

    def test_double(self):
        rv = self.obj.emit("test-double", 1.234)
        self.assertEqual(rv, 1.234)

    def test_int64(self):
        rv = self.obj.emit("test-int64", 102030405)
        self.assertEqual(rv, 102030405)

        rv = self.obj.emit("test-int64", GLib.MAXINT64)
        self.assertEqual(rv, GLib.MAXINT64 - 1)

        rv = self.obj.emit("test-int64", GLib.MININT64)
        self.assertEqual(rv, GLib.MININT64)

    def test_string(self):
        rv = self.obj.emit("test-string", "str")
        self.assertEqual(rv, "str")

    def test_object(self):
        rv = self.obj.emit("test-object", self)
        self.assertEqual(rv, self)

    def test_paramspec(self):
        rv = self.obj.emit("test-paramspec")
        self.assertEqual(rv.name, "test-param")
        self.assertEqual(rv.nick, "test")

    def test_paramspec_in(self):
        rv = GObject.param_spec_boolean(
            "mybool", "test-bool", "do something", True, GObject.ParamFlags.READABLE
        )

        rv2 = self.obj.emit("test-paramspec-in", rv)
        self.assertEqual(type(rv), type(rv2))
        self.assertEqual(rv2.name, "mybool")
        self.assertEqual(rv2.nick, "test-bool")

    def test_C_paramspec(self):
        self.notify_called = False

        def cb_notify(obj, prop):
            self.notify_called = True
            self.assertEqual(obj, self.obj)
            self.assertEqual(prop.name, "testprop")

        self.obj.connect("notify", cb_notify)
        self.obj.set_property("testprop", 42)
        self.assertTrue(self.notify_called)

    def test_gvalue(self):
        # implicit int
        rv = self.obj.emit("test-gvalue", 42)
        self.assertEqual(rv, 42)

        # explicit float
        v = GObject.Value(GObject.TYPE_FLOAT, 1.234)
        rv = self.obj.emit("test-gvalue", v)
        self.assertAlmostEqual(rv, 1.234, places=4)

        # implicit float
        rv = self.obj.emit("test-gvalue", 1.234)
        self.assertAlmostEqual(rv, 1.234, places=4)

        # explicit int64
        v = GObject.Value(GObject.TYPE_INT64, GLib.MAXINT64)
        rv = self.obj.emit("test-gvalue", v)
        self.assertEqual(rv, GLib.MAXINT64)

        # explicit uint64
        v = GObject.Value(GObject.TYPE_UINT64, GLib.MAXUINT64)
        rv = self.obj.emit("test-gvalue", v)
        self.assertEqual(rv, GLib.MAXUINT64)

    @unittest.expectedFailure  # https://bugzilla.gnome.org/show_bug.cgi?id=705291
    def test_gvalue_implicit_int64(self):
        # implicit int64
        rv = self.obj.emit("test-gvalue", GLib.MAXINT64)
        self.assertEqual(rv, GLib.MAXINT64)

        # implicit uint64
        rv = self.obj.emit("test-gvalue", GLib.MAXUINT64)
        self.assertEqual(rv, GLib.MAXUINT64)

    def test_gvalue_ret(self):
        self.assertEqual(
            self.obj.emit("test-gvalue-ret", GObject.TYPE_INT), GLib.MAXINT
        )
        self.assertEqual(
            self.obj.emit("test-gvalue-ret", GObject.TYPE_UINT), GLib.MAXUINT
        )
        self.assertEqual(
            self.obj.emit("test-gvalue-ret", GObject.TYPE_INT64), GLib.MAXINT64
        )
        self.assertEqual(
            self.obj.emit("test-gvalue-ret", GObject.TYPE_UINT64), GLib.MAXUINT64
        )
        self.assertEqual(self.obj.emit("test-gvalue-ret", GObject.TYPE_STRING), "hello")


class TestCMarshaller(_TestCMarshaller, unittest.TestCase):
    pass


# Test for 374653


class TestPyGValue(unittest.TestCase):
    def test_none_null_boxed_conversion(self):
        class C(GObject.GObject):
            __gsignals__ = {
                "my_boxed_signal": (GObject.SignalFlags.RUN_LAST, GObject.TYPE_STRV, ())
            }

        obj = C()
        obj.connect("my-boxed-signal", lambda obj: None)
        sys.last_type = None
        obj.emit("my-boxed-signal")
        assert not sys.last_type


class TestSignalDecorator(unittest.TestCase):
    class Decorated(GObject.GObject):
        value = 0

        @GObject.Signal
        def pushed(self):
            """This will push."""
            self.value += 1

        @GObject.Signal(flags=GObject.SignalFlags.RUN_LAST)
        def pulled(self):
            self.value -= 1

        @GObject.Signal(flags=GObject.SignalFlags.DETAILED)
        def detailed(self):
            self.value -= 1

        stomped = GObject.Signal("stomped", arg_types=(int,), doc="This will stomp.")
        unnamed = GObject.Signal()

    class DecoratedOverride(GObject.GObject):
        overridden_closure_called = False
        notify_called = False
        value = GObject.Property(type=int, default=0)

        @GObject.SignalOverride
        def notify(self, *args, **kargs):
            self.overridden_closure_called = True

        def on_notify(self, obj, prop):
            self.notify_called = True

    def setUp(self):
        self.unnamedCalled = False

    def onUnnamed(self, obj):
        self.unnamedCalled = True

    def test_disconnect(self):
        decorated = self.Decorated()
        id_ = decorated.pushed.connect(lambda *args: None)
        decorated.pushed.disconnect(id_)

    def test_signal_repr(self):
        decorated = self.Decorated()
        assert repr(decorated.pushed) == 'BoundSignal("pushed")'

    def test_signal_call(self):
        decorated = self.Decorated()
        assert decorated.value == 0
        decorated.pushed()
        assert decorated.value == 1

    def test_connect_detailed(self):
        decorated = self.Decorated()
        id_ = decorated.detailed.connect_detailed(lambda *args: None, "foo")
        decorated.pushed.disconnect(id_)

    def test_get_signal_args(self):
        self.assertEqual(
            self.Decorated.pushed.get_signal_args(),
            (GObject.SignalFlags.RUN_FIRST, None, (), None, None),
        )
        self.assertEqual(
            self.Decorated.pulled.get_signal_args(),
            (GObject.SignalFlags.RUN_LAST, None, (), None, None),
        )
        self.assertEqual(
            self.Decorated.stomped.get_signal_args(),
            (GObject.SignalFlags.RUN_FIRST, None, (int,), None, None),
        )

    def test_closures_called(self):
        decorated = self.Decorated()
        self.assertEqual(decorated.value, 0)
        decorated.pushed.emit()
        self.assertEqual(decorated.value, 1)
        decorated.pulled.emit()
        self.assertEqual(decorated.value, 0)

    def test_signal_copy(self):
        blah = self.Decorated.stomped.copy("blah")
        self.assertEqual(str(blah), blah)
        self.assertEqual(blah.func, self.Decorated.stomped.func)
        self.assertEqual(blah.flags, self.Decorated.stomped.flags)
        self.assertEqual(blah.return_type, self.Decorated.stomped.return_type)
        self.assertEqual(blah.arg_types, self.Decorated.stomped.arg_types)
        self.assertEqual(blah.__doc__, self.Decorated.stomped.__doc__)

    def test_doc_string(self):
        # Test the two techniques for setting doc strings on the signals
        # class variables, through the "doc" keyword or as the getter doc string.
        self.assertEqual(self.Decorated.stomped.__doc__, "This will stomp.")
        self.assertEqual(self.Decorated.pushed.__doc__, "This will push.")

    def test_unnamed_signal_gets_named(self):
        self.assertEqual(str(self.Decorated.unnamed), "unnamed")

    def test_unnamed_signal_gets_called(self):
        obj = self.Decorated()
        obj.connect("unnamed", self.onUnnamed)
        self.assertEqual(self.unnamedCalled, False)
        obj.emit("unnamed")
        self.assertEqual(self.unnamedCalled, True)

    def test_overridden_signal(self):
        # Test that the pushed signal is called in with super and the override
        # which should both increment the "value" to 3
        obj = self.DecoratedOverride()
        obj.connect("notify", obj.on_notify)
        self.assertEqual(obj.value, 0)
        obj.value = 1
        self.assertEqual(obj.value, 1)
        self.assertTrue(obj.overridden_closure_called)
        self.assertTrue(obj.notify_called)


class TestSignalConnectors(unittest.TestCase):
    class CustomButton(GObject.GObject):
        on_notify_called = False
        value = GObject.Property(type=int)

        @GObject.Signal(arg_types=(int,))
        def clicked(self, value):
            self.value = value

    def setUp(self):
        self.obj = None
        self.value = None

    def on_clicked(self, obj, value):
        self.obj = obj
        self.value = value

    def test_signal_notify(self):
        def on_notify(obj, param):
            obj.on_notify_called = True

        obj = self.CustomButton()
        obj.connect("notify", on_notify)
        self.assertFalse(obj.on_notify_called)
        obj.notify("value")
        self.assertTrue(obj.on_notify_called)

    def test_signal_emit(self):
        # standard callback connection with different forms of emit.
        obj = self.CustomButton()
        obj.connect("clicked", self.on_clicked)

        # vanilla
        obj.emit("clicked", 1)
        self.assertEqual(obj.value, 1)
        self.assertEqual(obj, self.obj)
        self.assertEqual(self.value, 1)

        # using class signal as param
        self.obj = None
        self.value = None
        obj.emit(self.CustomButton.clicked, 1)
        self.assertEqual(obj, self.obj)
        self.assertEqual(self.value, 1)

        # using bound signal as param
        self.obj = None
        self.value = None
        obj.emit(obj.clicked, 1)
        self.assertEqual(obj, self.obj)
        self.assertEqual(self.value, 1)

        # using bound signal with emit
        self.obj = None
        self.value = None
        obj.clicked.emit(1)
        self.assertEqual(obj, self.obj)
        self.assertEqual(self.value, 1)

    def test_signal_class_connect(self):
        obj = self.CustomButton()
        obj.connect(self.CustomButton.clicked, self.on_clicked)
        obj.emit("clicked", 2)
        self.assertEqual(obj, self.obj)
        self.assertEqual(self.value, 2)

    def test_signal_bound_connect(self):
        obj = self.CustomButton()
        obj.clicked.connect(self.on_clicked)
        obj.emit("clicked", 3)
        self.assertEqual(obj, self.obj)
        self.assertEqual(self.value, 3)


class _ConnectDataTestBase:
    # Notes:
    #  - self.Object is overridden in sub-classes.
    #  - Numeric suffixes indicate the number of user data args passed in.
    Object = None

    def run_connect_test(self, emit_args, user_data, flags=0):
        obj = self.Object()
        callback_args = []

        def callback(*args):
            callback_args.append(args)
            return 0

        obj.connect_data(
            "sig-with-int64-prop", callback, connect_flags=flags, *user_data
        )
        obj.emit("sig-with-int64-prop", *emit_args)
        self.assertEqual(len(callback_args), 1)
        return callback_args[0]

    def test_0(self):
        obj, value = self.run_connect_test([GLib.MAXINT64], user_data=[])
        self.assertIsInstance(obj, self.Object)
        self.assertEqual(value, GLib.MAXINT64)

    def test_1(self):
        obj, value, data = self.run_connect_test([GLib.MAXINT64], user_data=["mydata"])
        self.assertIsInstance(obj, self.Object)
        self.assertEqual(value, GLib.MAXINT64)
        self.assertEqual(data, "mydata")

    def test_after_0(self):
        obj, value = self.run_connect_test(
            [GLib.MAXINT64], user_data=[], flags=GObject.ConnectFlags.AFTER
        )
        self.assertIsInstance(obj, self.Object)
        self.assertEqual(value, GLib.MAXINT64)

    def test_after_1(self):
        obj, value, data = self.run_connect_test(
            [GLib.MAXINT64], user_data=["mydata"], flags=GObject.ConnectFlags.AFTER
        )
        self.assertIsInstance(obj, self.Object)
        self.assertEqual(value, GLib.MAXINT64)
        self.assertEqual(data, "mydata")

    def test_swaped_0(self):
        # Swapped only works with a single user data argument.
        with self.assertRaises(ValueError):
            self.run_connect_test(
                [GLib.MAXINT64], user_data=[], flags=GObject.ConnectFlags.SWAPPED
            )

    def test_swaped_1(self):
        # Notice obj and data are reversed in the return.
        data, value, obj = self.run_connect_test(
            [GLib.MAXINT64], user_data=["mydata"], flags=GObject.ConnectFlags.SWAPPED
        )
        self.assertIsInstance(obj, self.Object)
        self.assertEqual(value, GLib.MAXINT64)
        self.assertEqual(data, "mydata")

    def test_swaped_2(self):
        # Swapped only works with a single user data argument.
        with self.assertRaises(ValueError):
            self.run_connect_test(
                [GLib.MAXINT64], user_data=[1, 2], flags=GObject.ConnectFlags.SWAPPED
            )

    def test_after_and_swapped_0(self):
        # Swapped only works with a single user data argument.
        with self.assertRaises(ValueError):
            self.run_connect_test(
                [GLib.MAXINT64],
                user_data=[],
                flags=GObject.ConnectFlags.AFTER | GObject.ConnectFlags.SWAPPED,
            )

    def test_after_and_swapped_1(self):
        # Notice obj and data are reversed in the return.
        data, value, obj = self.run_connect_test(
            [GLib.MAXINT64],
            user_data=["mydata"],
            flags=GObject.ConnectFlags.AFTER | GObject.ConnectFlags.SWAPPED,
        )
        self.assertIsInstance(obj, self.Object)
        self.assertEqual(value, GLib.MAXINT64)
        self.assertEqual(data, "mydata")

    def test_after_and_swapped_2(self):
        # Swapped only works with a single user data argument.
        with self.assertRaises(ValueError):
            self.run_connect_test(
                [GLib.MAXINT64],
                user_data=[],
                flags=GObject.ConnectFlags.AFTER | GObject.ConnectFlags.SWAPPED,
            )


class TestConnectDataNonIntrospected(unittest.TestCase, _ConnectDataTestBase):
    # This tests connect_data with non-introspected signals
    # (created in Python in this case).
    class Object(GObject.Object):
        test = GObject.Signal()
        sig_with_int64_prop = GObject.Signal(
            return_type=GObject.TYPE_INT64,
            arg_types=[GObject.TYPE_INT64],
            flags=GObject.SignalFlags.RUN_LAST,
        )


class TestConnectDataIntrospected(unittest.TestCase, _ConnectDataTestBase):
    # This tests connect_data with introspected signals brought in from Regress.
    Object = Regress.TestObj


class TestInstallSignals(unittest.TestCase):
    # These tests only test how signalhelper.install_signals works
    # with the __gsignals__ dict and therefore does not need to use
    # GObject as a base class because that would automatically call
    # install_signals within the meta-class.
    class Base:
        __gsignals__ = {"test": (0, None, ())}

    class Sub1(Base):
        pass

    class Sub2(Base):
        @GObject.Signal
        def sub2test(self):
            pass

    def setUp(self):
        self.assertEqual(len(self.Base.__gsignals__), 1)
        signalhelper.install_signals(self.Base)
        self.assertEqual(len(self.Base.__gsignals__), 1)

    def test_subclass_gets_empty_gsignals_dict(self):
        # Installing signals will add the __gsignals__ dict to a class
        # if it doesn't already exists.
        self.assertFalse("__gsignals__" in self.Sub1.__dict__)
        signalhelper.install_signals(self.Sub1)
        self.assertTrue("__gsignals__" in self.Sub1.__dict__)
        # Sub1 should only contain an empty signals dict, this tests:
        # https://bugzilla.gnome.org/show_bug.cgi?id=686496
        self.assertEqual(self.Sub1.__dict__["__gsignals__"], {})

    def test_subclass_with_decorator_gets_gsignals_dict(self):
        self.assertFalse("__gsignals__" in self.Sub2.__dict__)
        signalhelper.install_signals(self.Sub2)
        self.assertTrue("__gsignals__" in self.Sub2.__dict__)
        self.assertEqual(len(self.Base.__gsignals__), 1)
        self.assertEqual(len(self.Sub2.__gsignals__), 1)
        self.assertTrue("sub2test" in self.Sub2.__gsignals__)

        # Make sure the vfunc was added
        self.assertTrue(hasattr(self.Sub2, "do_sub2test"))


class TestPython3Signals(unittest.TestCase):
    class AnnotatedClass(GObject.GObject):
        @GObject.Signal
        def sig1(self, a: int, b: float):
            pass

        @GObject.Signal(flags=GObject.SignalFlags.RUN_LAST)
        def sig2_with_return(self, a: int, b: float) -> str:
            return "test"

    def test_annotations(self):
        self.assertEqual(
            signalhelper.get_signal_annotations(self.AnnotatedClass.sig1.func),
            (None, (int, float)),
        )
        self.assertEqual(
            signalhelper.get_signal_annotations(
                self.AnnotatedClass.sig2_with_return.func
            ),
            (str, (int, float)),
        )

        self.assertEqual(
            self.AnnotatedClass.sig2_with_return.get_signal_args(),
            (GObject.SignalFlags.RUN_LAST, str, (int, float), None, None),
        )
        self.assertEqual(self.AnnotatedClass.sig2_with_return.arg_types, (int, float))
        self.assertEqual(self.AnnotatedClass.sig2_with_return.return_type, str)

    def test_emit_return(self):
        obj = self.AnnotatedClass()
        self.assertEqual(obj.sig2_with_return.emit(1, 2.0), "test")


class TestSignalModuleLevelFunctions(unittest.TestCase):
    def test_signal_list_ids_with_invalid_type(self):
        with self.assertRaisesRegex(
            TypeError, "type must be instantiable or an interface.*"
        ):
            GObject.signal_list_ids(GObject.TYPE_INVALID)

    def test_signal_list_ids(self):
        with self.assertRaisesRegex(
            TypeError, "type must be instantiable or an interface.*"
        ):
            GObject.signal_list_ids(GObject.TYPE_INT)

        ids = GObject.signal_list_ids(C)
        self.assertEqual(len(ids), 1)
        # Note canonicalized names
        self.assertEqual(GObject.signal_name(ids[0]), "my-signal")
        # There is no signal 0 in gobject
        self.assertEqual(GObject.signal_name(0), None)

    def test_signal_lookup_with_invalid_type(self):
        with self.assertRaisesRegex(
            TypeError, "type must be instantiable or an interface.*"
        ):
            GObject.signal_lookup("NOT_A_SIGNAL_NAME", GObject.TYPE_INVALID)

    def test_signal_lookup(self):
        ids = GObject.signal_list_ids(C)
        self.assertEqual(ids[0], GObject.signal_lookup("my_signal", C))
        self.assertEqual(ids[0], GObject.signal_lookup("my-signal", C))

        with self.assertRaisesRegex(
            TypeError, "type must be instantiable or an interface.*"
        ):
            GObject.signal_lookup("NOT_A_SIGNAL_NAME", GObject.TYPE_INT)

        # Invalid signal names return 0 instead of raising
        self.assertEqual(GObject.signal_lookup("NOT_A_SIGNAL_NAME", C), 0)

    def test_signal_query(self):
        (my_signal_id,) = GObject.signal_list_ids(C)

        # Form is: (id, name, gtype, arg_count, return_type, (arg_type1, ...))
        my_signal_expected_query_result = [
            my_signal_id,
            "my-signal",
            C.__gtype__,
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (GObject.TYPE_INT,),
        ]
        # signal_query(name, type)
        self.assertEqual(
            list(GObject.signal_query("my-signal", C)), my_signal_expected_query_result
        )
        # signal_query(signal_id)
        self.assertEqual(
            list(GObject.signal_query(my_signal_id)), my_signal_expected_query_result
        )
        # invalid query returns None instead of raising
        self.assertEqual(GObject.signal_query(0), None)
        self.assertEqual(GObject.signal_query("NOT_A_SIGNAL", C), None)


class TestIntrospectedSignals(unittest.TestCase):
    def test_object_param_signal(self):
        obj = Regress.TestObj()

        def callback(obj, obj_param):
            self.assertEqual(obj_param.props.int, 3)
            self.assertGreater(obj_param.__grefcount__, 1)
            obj.called = True

        obj.called = False
        obj.connect("sig-with-obj", callback)
        obj.emit_sig_with_obj()
        self.assertTrue(obj.called)

    def test_connect_after(self):
        obj = Regress.TestObj()

        def callback(obj, obj_param):
            obj.called = True

        obj.called = False
        obj.connect_after("sig-with-obj", callback)
        obj.emit_sig_with_obj()
        self.assertTrue(obj.called)

    def test_int64_param_from_py(self):
        obj = Regress.TestObj()

        def callback(obj, i):
            obj.callback_i = i
            return i

        obj.callback_i = None
        obj.connect("sig-with-int64-prop", callback)
        rv = obj.emit("sig-with-int64-prop", GLib.MAXINT64)
        self.assertEqual(rv, GLib.MAXINT64)
        self.assertEqual(obj.callback_i, GLib.MAXINT64)

    def test_uint64_param_from_py(self):
        obj = Regress.TestObj()

        def callback(obj, i):
            obj.callback_i = i
            return i

        obj.callback_i = None
        obj.connect("sig-with-uint64-prop", callback)
        rv = obj.emit("sig-with-uint64-prop", GLib.MAXUINT64)
        self.assertEqual(rv, GLib.MAXUINT64)
        self.assertEqual(obj.callback_i, GLib.MAXUINT64)

    def test_int64_param_from_c(self):
        obj = Regress.TestObj()

        def callback(obj, i):
            obj.callback_i = i
            return i

        obj.callback_i = None

        obj.connect("sig-with-int64-prop", callback)
        obj.emit_sig_with_int64()
        self.assertEqual(obj.callback_i, GLib.MAXINT64)

    def test_uint64_param_from_c(self):
        obj = Regress.TestObj()

        def callback(obj, i):
            obj.callback_i = i
            return i

        obj.callback_i = None

        obj.connect("sig-with-uint64-prop", callback)
        obj.emit_sig_with_uint64()
        self.assertEqual(obj.callback_i, GLib.MAXUINT64)

    def test_intarray_ret(self):
        obj = Regress.TestObj()

        def callback(obj, i):
            obj.callback_i = i
            return [i, i + 1]

        obj.callback_i = None

        try:
            obj.connect("sig-with-intarray-ret", callback)
        except TypeError as e:
            # compat with g-i 1.34.x
            if "unknown signal" in str(e):
                return
            raise

        rv = obj.emit("sig-with-intarray-ret", 42)
        self.assertEqual(obj.callback_i, 42)
        self.assertEqual(type(rv), GLib.Array)
        self.assertEqual(rv.len, 2)

    @unittest.skip("https://bugzilla.gnome.org/show_bug.cgi?id=669496")
    def test_array_parm(self):
        obj = Regress.TestObj()

        def callback(obj, arr):
            obj.callback_arr = arr

        obj.connect("sig-with-array-prop", callback)
        obj.callback_arr = None
        self.assertEqual(obj.emit("sig-with-array-prop", [1, 2, GLib.MAXUINT]), None)
        self.assertEqual(obj.callback_arr, [1, 2, GLib.MAXUINT])

    def test_held_struct_ref(self):
        held_structs = []

        def callback(obj, struct):
            # The struct held by Python will become a copy after this callback exits.
            struct.some_int = 42
            struct.some_int8 = 42
            held_structs.append(struct)

        struct = Regress.TestSimpleBoxedA()
        obj = Regress.TestObj()

        self.assertEqual(struct.some_int, 0)
        self.assertEqual(struct.some_int8, 0)

        obj.connect("test-with-static-scope-arg", callback)
        obj.emit("test-with-static-scope-arg", struct)

        # The held struct will be a copy of the modified struct.
        self.assertEqual(len(held_structs), 1)
        held_struct = held_structs[0]
        self.assertEqual(held_struct.some_int, 42)
        self.assertEqual(held_struct.some_int8, 42)

        # Boxed equality checks pointers by default.
        self.assertNotEqual(struct, held_struct)

    def test_action(self):
        obj = Regress.TestAction()
        other_obj = obj.emit("action")
        self.assertEqual(other_obj.__grefcount__, 1)
        other_obj2 = obj.emit("action2")
        self.assertIsNone(other_obj2)


class TestIntrospectedSignalsIssue158(unittest.TestCase):
    """The test for https://gitlab.gnome.org/GNOME/pygobject/issues/158."""

    _obj_sig_names = [
        sig.get_name() for sig in repo.find_by_name("Regress", "TestObj").get_signals()
    ]

    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)
        self._gc_thread_stop = False

    def _gc_thread(self):
        while not self._gc_thread_stop:
            gc.collect()
            time.sleep(0.010)

    def _callback(self, *args):
        pass

    def test_run(self):
        """Manually trigger GC from a different thread periodicaly
        while the main thread keeps connecting/disconnecting to/from signals.

        It takes a lot of time to reproduce the issue. It is possible to make it
        fail reliably by changing the code of pygobject_unwatch_closure slightly from:
          PyGObjectData *inst_data = data;
          inst_data->closures = g_slist_remove (inst_data->closures, closure);
        to
          PyGObjectData *inst_data = data;
          GSList *tmp = g_slist_remove (inst_data->closures, closure);
          g_usleep(G_USEC_PER_SEC/10);
          inst_data->closures = tmp;
        """
        obj = Regress.TestObj()
        gc_thread = threading.Thread(target=self._gc_thread)
        gc_thread.start()

        for _ in range(8):
            handlers = [obj.connect(sig, self._callback) for sig in self._obj_sig_names]
            time.sleep(0.010)
            while len(handlers) > 0:
                obj.disconnect(handlers.pop())

        self._gc_thread_stop = True
        gc_thread.join()


class _ConnectObjectTestBase:
    # Notes:
    #  - self.Object is overridden in sub-classes.
    #  - Numeric suffixes indicate the number of user data args passed in.
    Object = None
    SwapObject = None

    def run_connect_test(self, emit_args, user_data, flags=0):
        obj = self.Object()
        callback_args = []
        swap_obj = self.SwapObject()

        def callback(*args):
            callback_args.append(args)
            return 0

        if flags & GObject.ConnectFlags.AFTER:
            connect_func = obj.connect_object_after
        else:
            connect_func = obj.connect_object

        with capture_gi_deprecation_warnings():
            connect_func("sig-with-int64-prop", callback, swap_obj, *user_data)
        obj.emit("sig-with-int64-prop", *emit_args)
        self.assertEqual(len(callback_args), 1)
        return callback_args[0]

    def test_0(self):
        obj, value = self.run_connect_test([GLib.MAXINT64], user_data=[])
        self.assertIsInstance(obj, self.SwapObject)
        self.assertEqual(value, GLib.MAXINT64)

    def test_1(self):
        obj, value, data = self.run_connect_test([GLib.MAXINT64], user_data=["mydata"])
        self.assertIsInstance(obj, self.SwapObject)
        self.assertEqual(value, GLib.MAXINT64)
        self.assertEqual(data, "mydata")

    def test_2(self):
        obj, value, data1, data2 = self.run_connect_test(
            [GLib.MAXINT64], user_data=["mydata1", "mydata2"]
        )
        self.assertIsInstance(obj, self.SwapObject)
        self.assertEqual(value, GLib.MAXINT64)
        self.assertEqual(data1, "mydata1")
        self.assertEqual(data2, "mydata2")

    def test_after_0(self):
        obj, value = self.run_connect_test(
            [GLib.MAXINT64], user_data=[], flags=GObject.ConnectFlags.AFTER
        )
        self.assertIsInstance(obj, self.SwapObject)
        self.assertEqual(value, GLib.MAXINT64)

    def test_after_1(self):
        obj, value, data = self.run_connect_test(
            [GLib.MAXINT64], user_data=["mydata"], flags=GObject.ConnectFlags.AFTER
        )
        self.assertIsInstance(obj, self.SwapObject)
        self.assertEqual(value, GLib.MAXINT64)
        self.assertEqual(data, "mydata")

    def test_after_2(self):
        obj, value, data1, data2 = self.run_connect_test(
            [GLib.MAXINT64],
            user_data=["mydata1", "mydata2"],
            flags=GObject.ConnectFlags.AFTER,
        )
        self.assertIsInstance(obj, self.SwapObject)
        self.assertEqual(value, GLib.MAXINT64)
        self.assertEqual(data1, "mydata1")
        self.assertEqual(data2, "mydata2")


class TestConnectGObjectNonIntrospected(unittest.TestCase, _ConnectObjectTestBase):
    # This tests connect_object with non-introspected signals
    # (created in Python in this case).
    class Object(GObject.Object):
        test = GObject.Signal()
        sig_with_int64_prop = GObject.Signal(
            return_type=GObject.TYPE_INT64,
            arg_types=[GObject.TYPE_INT64],
            flags=GObject.SignalFlags.RUN_LAST,
        )

    # Object passed for swapping is GObject based.
    class SwapObject(GObject.Object):
        pass


class TestConnectGObjectIntrospected(unittest.TestCase, _ConnectObjectTestBase):
    # This tests connect_object with introspected signals brought in from Regress.
    Object = Regress.TestObj

    # Object passed for swapping is GObject based.
    class SwapObject(GObject.Object):
        pass


class TestConnectPyObjectNonIntrospected(unittest.TestCase, _ConnectObjectTestBase):
    # This tests connect_object with non-introspected signals
    # (created in Python in this case).
    class Object(GObject.Object):
        test = GObject.Signal()
        sig_with_int64_prop = GObject.Signal(
            return_type=GObject.TYPE_INT64,
            arg_types=[GObject.TYPE_INT64],
            flags=GObject.SignalFlags.RUN_LAST,
        )

    # Object passed for swapping is pure Python
    SwapObject = object


class TestConnectPyObjectIntrospected(unittest.TestCase, _ConnectObjectTestBase):
    # This tests connect_object with introspected signals brought in from Regress.
    Object = Regress.TestObj

    # Object passed for swapping is pure Python
    SwapObject = object


class _RefCountTestBase:
    # NOTE: ref counts are always one more than expected because the getrefcount()
    # function adds a ref for the input argument.

    # Sub-classes set this
    Object = None

    class PyData:
        pass

    @unittest.skipUnless(hasattr(sys, "getrefcount"), "no sys.getrefcount")
    def test_callback_ref_count_del(self):
        def callback(obj, value):
            return value // 2

        callback_ref = weakref.ref(callback)
        self.assertEqual(
            sys.getrefcount(callback), 2 if sys.version_info < (3, 14) else 1
        )

        obj = self.Object()
        obj.connect("sig-with-int64-prop", callback)
        self.assertEqual(
            sys.getrefcount(callback), 3 if sys.version_info < (3, 14) else 2
        )

        del callback
        self.assertEqual(sys.getrefcount(callback_ref()), 2)

        res = obj.emit("sig-with-int64-prop", 42)
        self.assertEqual(res, 21)
        self.assertEqual(
            sys.getrefcount(callback_ref), 2 if sys.version_info < (3, 14) else 1
        )

        del obj
        self.assertIsNone(callback_ref())

    @unittest.skipUnless(hasattr(sys, "getrefcount"), "no sys.getrefcount")
    def test_callback_ref_count_disconnect(self):
        def callback(obj, value):
            return value // 2

        callback_ref = weakref.ref(callback)
        self.assertEqual(
            sys.getrefcount(callback), 2 if sys.version_info < (3, 14) else 1
        )

        obj = self.Object()
        handler_id = obj.connect("sig-with-int64-prop", callback)
        self.assertEqual(
            sys.getrefcount(callback), 3 if sys.version_info < (3, 14) else 2
        )

        del callback
        self.assertEqual(sys.getrefcount(callback_ref()), 2)

        res = obj.emit("sig-with-int64-prop", 42)
        self.assertEqual(res, 21)
        self.assertEqual(
            sys.getrefcount(callback_ref), 2 if sys.version_info < (3, 14) else 1
        )

        obj.disconnect(handler_id)
        self.assertIsNone(callback_ref())

    @unittest.skipUnless(hasattr(sys, "getrefcount"), "no sys.getrefcount")
    def test_callback_ref_count_disconnect_by_func(self):
        def callback(obj, value):
            return value // 2

        callback_ref = weakref.ref(callback)
        self.assertEqual(
            sys.getrefcount(callback), 2 if sys.version_info < (3, 14) else 1
        )

        obj = self.Object()
        obj.connect("sig-with-int64-prop", callback)
        self.assertEqual(
            sys.getrefcount(callback), 3 if sys.version_info < (3, 14) else 2
        )

        del callback
        self.assertEqual(sys.getrefcount(callback_ref()), 2)

        res = obj.emit("sig-with-int64-prop", 42)
        self.assertEqual(res, 21)
        self.assertEqual(
            sys.getrefcount(callback_ref), 2 if sys.version_info < (3, 14) else 1
        )

        obj.disconnect_by_func(callback_ref())
        self.assertIsNone(callback_ref())

    @unittest.skipUnless(hasattr(sys, "getrefcount"), "no sys.getrefcount")
    def test_user_data_ref_count(self):
        def callback(obj, value, data):
            return value // 2

        data = self.PyData()
        data_ref = weakref.ref(data)
        self.assertEqual(sys.getrefcount(data), 2 if sys.version_info < (3, 14) else 1)

        obj = self.Object()
        obj.connect("sig-with-int64-prop", callback, data)
        self.assertEqual(sys.getrefcount(data), 3 if sys.version_info < (3, 14) else 2)

        del data
        self.assertEqual(sys.getrefcount(data_ref()), 2)

        res = obj.emit("sig-with-int64-prop", 42)
        self.assertEqual(res, 21)
        self.assertEqual(sys.getrefcount(data_ref()), 2)

        del obj
        self.assertIsNone(data_ref())

    @unittest.skipUnless(hasattr(sys, "getrefcount"), "no sys.getrefcount")
    @unittest.expectedFailure  # https://bugzilla.gnome.org/show_bug.cgi?id=688064
    def test_object_ref_count(self):
        # connect_object() should only weakly reference the object passed in
        # and auto-disconnect the signal when the object is destroyed.
        def callback(data, value):
            return value // 2

        data = GObject.Object()
        data_ref = weakref.ref(data)
        self.assertEqual(sys.getrefcount(data), 2)

        obj = self.Object()
        handler_id = obj.connect_object("sig-with-int64-prop", callback, data)
        self.assertEqual(sys.getrefcount(data), 2)

        res = obj.emit("sig-with-int64-prop", 42)
        self.assertEqual(res, 21)
        self.assertEqual(sys.getrefcount(data), 2)

        del data

        self.assertIsNone(data_ref())
        self.assertFalse(obj.handler_is_connected(handler_id))


class TestRefCountsNonIntrospected(unittest.TestCase, _RefCountTestBase):
    class Object(GObject.Object):
        sig_with_int64_prop = GObject.Signal(
            return_type=GObject.TYPE_INT64,
            arg_types=[GObject.TYPE_INT64],
            flags=GObject.SignalFlags.RUN_LAST,
        )


class TestRefCountsIntrospected(unittest.TestCase, _RefCountTestBase):
    Object = Regress.TestObj


class TestClosureRefCycle(unittest.TestCase):
    def test_closure_ref_cycle_unreachable(self):
        # https://bugzilla.gnome.org/show_bug.cgi?id=731501

        called = []

        def on_add(store, *args):
            called.append(store)

        store = Gio.ListStore()
        store.connect_object("items-changed", on_add, store)

        # Remove all Python references to the object and keep it alive
        # on the C level.
        x = Gio.FileInfo()
        x.set_attribute_object("store", store)
        del store
        gc.collect()

        # get it back and trigger the signal
        x.get_attribute_object("store").append(Gio.FileInfo())

        self.assertEqual(len(called), 1)
        self.assertTrue(called[0].__grefcount__ > 0)



import unittest
import ctypes
from ctypes import c_void_p, py_object, c_char_p

import gi
from gi.repository import Gio


def get_capi():
    if not hasattr(ctypes, "pythonapi"):
        return None

    class CAPI(ctypes.Structure):
        _fields_ = [
            ("", c_void_p),
            ("", c_void_p),
            ("", c_void_p),
            ("newgobj", ctypes.PYFUNCTYPE(py_object, c_void_p)),
        ]

    api_obj = gi._gobject._PyGObject_API
    func_type = ctypes.PYFUNCTYPE(c_void_p, py_object, c_char_p)
    PyCapsule_GetPointer = func_type(("PyCapsule_GetPointer", ctypes.pythonapi))
    ptr = PyCapsule_GetPointer(api_obj, b"gobject._PyGObject_API")

    ptr = ctypes.cast(ptr, ctypes.POINTER(CAPI))
    return ptr.contents


API = get_capi()


@unittest.skipUnless(API, "no pythonapi support")
class TestPythonCAPI(unittest.TestCase):
    def test_newgobj(self):
        w = Gio.FileInfo()
        # XXX: ugh :/
        ptr = int(repr(w).split()[-1].split(")")[0], 16)

        capi = get_capi()
        new_w = capi.newgobj(ptr)
        assert w == new_w



import sys
import os
import re


def main(argv):
    # Fix paths in lcov files generated on a Windows host so they match our
    # current source layout.
    paths = argv[1:]

    for path in paths:
        print(f"cov-fixup: {path}")
        with open(path, encoding="utf-8") as h:
            text = h.read()

        text = text.replace("\\\\", "/").replace("\\", "/")
        new_root = os.getcwd()

        def make_abs(m):
            p = m.group(1)
            if p.startswith("C:/"):
                p = p.replace("C:/", "/c/")
            if not p.startswith("/"):
                p = os.path.join(new_root, p)
            return "SF:" + p

        text = re.sub(r"SF:(.*?)$", make_abs, text, 0, re.MULTILINE)

        candidate = None
        for old_root in set(re.findall(r":(.*?)/gi/.*?$", text, re.MULTILINE)):
            if candidate is None or len(old_root) < len(candidate):
                candidate = old_root

        if candidate:
            print(f"replacing {candidate} with {new_root}")
            text = text.replace(candidate, new_root)

        with open(path, "w", encoding="utf-8") as h:
            h.write(text)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



import gc
import sys
import warnings
import weakref

import pytest

from gi import PyGIWarning
from gi.repository import GObject, Gio, Regress


class DerivedObj(Regress.TestObj):
    def __init__(self):
        super().__init__()


@pytest.mark.parametrize("obj_type", [Regress.TestObj, DerivedObj])
def test_same_object_if_object_is_alive(obj_type):
    container = Regress.TestObj()
    obj = obj_type()
    container.set_bare(obj)

    assert container.props.bare is obj


@pytest.mark.parametrize("obj_type", [Regress.TestObj, DerivedObj])
def test_object_without_instance_data_gets_deleted(obj_type):
    container = Regress.TestObj()
    obj = obj_type()
    container.props.bare = obj

    ref = weakref.ref(obj)
    del obj
    gc.collect()  # for pypy

    assert container.props.bare
    assert ref() is None


@pytest.mark.parametrize("obj_type", [Regress.TestObj, DerivedObj])
def test_object_with_instance_data_gets_deleted(obj_type):
    container = Regress.TestObj()
    obj = obj_type()
    obj.instance_data = 1
    container.props.bare = obj

    ref = weakref.ref(obj)
    del obj
    gc.collect()  # for pypy

    assert ref() is None


@pytest.mark.parametrize("obj_type", [Regress.TestObj, DerivedObj])
def test_object_with_instance_data_retains_data(obj_type):
    container = Regress.TestObj()
    obj = obj_type()
    obj.instance_data = 1
    container.props.bare = obj

    del obj
    new_obj = container.props.bare

    assert new_obj.instance_data == 1


class ValueObj(Regress.TestObj):
    py_int = GObject.Property(type=int)

    py_model = GObject.Property(type=Gio.ListModel)

    def __init__(self, value=0):
        super().__init__()
        self.value = value


def test_value_object_retains_init_value():
    container = Regress.TestObj()
    obj = ValueObj(42)
    container.props.bare = obj

    del obj
    gc.collect()  # for pypy
    new_obj = container.props.bare

    assert new_obj.value == 42


def test_value_object_retains_property_value():
    container = Regress.TestObj()
    obj = ValueObj()
    obj.py_int = 42
    container.props.bare = obj

    del obj
    gc.collect()  # for pypy
    new_obj = container.props.bare

    assert new_obj.py_int == 42


class MySpecialListStore(GObject.Object, Gio.ListModel):
    def __init__(self):
        super().__init__()

    def do_get_item_type(self):
        return Regress.TestObj.__gtype__

    def do_get_n_items(self):
        return 0

    def do_get_item(self, n):
        raise NotImplementedError


def test_value_object_retains_object_property_value():
    container = Regress.TestObj()
    obj = ValueObj()
    obj.model = MySpecialListStore()
    container.props.bare = obj

    del obj
    gc.collect()  # for pypy
    new_obj = container.props.bare

    assert new_obj.model is not None


def test_object_with_property_binding():
    containers = [Regress.TestObj(), Regress.TestObj()]
    binder = Regress.TestObj()
    bindee = Regress.TestObj()

    binder.bind_property("bare", bindee, "bare", GObject.BindingFlags.SYNC_CREATE)
    containers[0].props.bare = binder
    containers[1].props.bare = bindee

    del binder
    del bindee

    gc.collect()  # for pypy

    new_binder = containers[0].props.bare
    new_bindee = containers[1].props.bare

    obj = Regress.TestObj()

    new_binder.props.bare = obj

    assert new_bindee.props.bare is obj


def test_object_with_signal_callback():
    container = Regress.TestObj()
    obj = ValueObj()

    def set_up_signal_handler(self):
        self.connect("first", lambda *_: self.set_property("int", 42))

    set_up_signal_handler(obj)

    container.props.bare = obj
    obj_id = id(obj)

    del obj
    gc.collect()  # for pypy
    new_obj = container.props.bare

    new_obj.emit("first")

    assert obj_id == id(new_obj)
    assert new_obj.props.int == 42


with warnings.catch_warnings():
    warnings.filterwarnings("ignore", ".*use __slots__.*")

    class SlotObj(Regress.TestObj):
        __slots__ = ["py_int"]

        def __init__(self):
            super().__init__()


def test_class_with_slots_raises_warning():
    with warnings.catch_warnings(record=True) as warn:

        class _SlotClass(GObject.Object):
            __slots__ = ["py_int"]

    assert warn[0].category is PyGIWarning
    assert "SlotClass shouldn't use __slots__." in str(warn[0].message)


def test_subclass_with_slots_raises_warning():
    class BaseClass(GObject.Object):
        pass

    with warnings.catch_warnings(record=True) as warn:

        class _SlotClass(BaseClass):
            __slots__ = ["py_int"]

    assert warn[0].category is PyGIWarning
    assert "SlotClass shouldn't use __slots__." in str(warn[0].message)


def test_slot_object_can_be_created():
    container = Regress.TestObj()
    obj = SlotObj()
    obj.py_int = 42
    container.props.bare = obj

    del obj
    gc.collect()  # for pypy
    new_obj = container.props.bare

    assert new_obj.py_int == 42


def test_slot_object_without_values_can_be_created():
    container = Regress.TestObj()
    obj = SlotObj()
    container.props.bare = obj

    del obj
    gc.collect()  # for pypy
    new_obj = container.props.bare

    assert new_obj


def test_objects_with_cyclic_dependency_and_instance_dict():
    # Test with a cycle:
    #
    # a --> b --> c --> d
    #       ^-----------'
    a, b, c, d = [Regress.TestObj(int=i) for i in range(4)]

    b.name = "b"

    a.props.bare = b
    b.props.bare = c
    b.c = c
    c.props.bare = d
    c.d = d
    d.props.bare = b
    d.b = b

    del b, c, d

    gc.collect()
    gc.collect()

    new_b = a.props.bare
    new_c = new_b.props.bare
    new_d = new_c.props.bare

    assert new_d.props.bare is new_b
    assert a.props.bare.name == "b"


def test_objects_with_cyclic_dependency_without_instance_dict():
    # Test with a cycle:
    #
    # a --> b --> c --> d
    #       ^-----------'
    a, b, c, d = [Regress.TestObj(int=i) for i in range(4)]

    a.props.bare = b
    b.props.bare = c
    c.props.bare = d
    d.props.bare = b

    del b, c, d

    gc.collect()
    gc.collect()

    new_b = a.props.bare
    new_c = new_b.props.bare
    new_d = new_c.props.bare

    assert new_d.props.bare is new_b


def test_objects_with_cyclic_dependency_and_instance_dict_no_content():
    # Test with a cycle:
    #
    # b <-> a <-> c
    #       ^---> d
    a, b, c, d = [Regress.TestObj(int=i) for i in range(4)]

    b.name = "b"

    a.props.bare = b
    a.all = [b, c, d]
    b.props.bare = a
    b.a = a
    c.props.bare = a
    c.a = a
    d.props.bare = a
    d.a = a
    del b, c, d

    gc.collect()
    gc.collect()

    assert a.props.bare.name == "b"


def test_chained_objects_are_collected():
    # a --> b --> c
    a, b, c = [Regress.TestObj(int=i) for i in range(3)]
    a.b = b
    b.c = c

    aref = a.weak_ref()
    cref = c.weak_ref()
    del a, b, c

    gc.collect()
    gc.collect()  # some more for PyPy
    gc.collect()
    gc.collect()

    assert aref() is None
    assert cref() is None


def test_objects_can_be_deleted():
    a = Regress.TestObj()

    pyref = weakref.ref(a)
    gref = a.weak_ref()
    del a

    gc.collect()  # for PyPy
    gc.collect()

    assert pyref() is None
    assert gref() is None


@pytest.mark.skipif(
    sys.implementation.name == "pypy", reason="Doesn't play nice with PyPy GC"
)
def test_gobject_cycle_is_collected():
    # Test with a cycle:
    #
    # a --> b --> c
    # ^-----------'
    a, b, c = [Regress.TestObj(int=i) for i in range(3)]

    a.b = b
    b.c = c
    c.a = a

    ref = a.weak_ref()
    del a, b, c

    gc.collect()
    gc.collect()

    assert ref() is None


def test_object_with_post_init():
    class PostInit(Regress.TestObj):
        number = GObject.Property(
            type=int, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT
        )

        def do_constructed(self):
            self.post_init_called = True
            assert self.number == 42

    obj = PostInit(number=42)

    assert obj.post_init_called
    assert obj.number == 42


def test_object_with_post_init_created_by_new():
    class PostInit(Regress.TestObj):
        number = GObject.Property(
            type=int, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT
        )

        def do_constructed(self):
            self.post_init_called = True
            assert self.number == 42

    obj = GObject.new(PostInit, number=42)

    assert obj.post_init_called
    assert obj.number == 42


class InitializationOrder(Regress.TestObj):
    def __init__(self):
        self.order = ["__init__1"]
        super().__init__()
        self.order.append("__init__2")

    def do_constructed(self):
        self.order.append("do_constructed")


def test_object_constructed_after_init():
    obj = InitializationOrder()

    assert obj.order == ["__init__1", "do_constructed", "__init__2"]


def test_object_constructed_after_init_by_new():
    obj = GObject.new(InitializationOrder)

    assert obj.order == ["__init__1", "__init__2", "do_constructed"]


def test_object_with_post_init_and_interface():
    class PostInit(Regress.TestObj, Regress.TestInterface):
        number = GObject.Property(type=int)

        def __init__(self):
            self.post_init_called = 0
            super().__init__()

        def do_constructed(self):
            super().do_constructed()
            self.post_init_called += 1

    class SubPostInit(PostInit):
        def do_constructed(self):
            super().do_constructed()

    obj = PostInit()
    subobj = SubPostInit()

    assert obj.post_init_called == 1
    assert subobj.post_init_called == 1


def test_object_with_post_init_raises_exception():
    class PostInit(Regress.TestObj):
        def do_constructed(self):
            raise ValueError("Catch me")

    with pytest.raises(ValueError, match="Catch me"):
        PostInit()
    with pytest.raises(ValueError, match="Catch me"):
        GObject.new(PostInit)



# Copyright (C) 2013 Simon Feltman <sfeltman@gnome.org>
#
#   test_repository.py: Test for the GIRepository module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA


import unittest
from collections import abc

import gi._gi as GIRepository
from gi.module import repository as repo
from gi.repository import GObject
from gi.repository import GIRepository as IntrospectedRepository


def find_child_info(info, getter_name, name):
    getter = getattr(info, getter_name)
    for child in getter():
        if child.get_name() == name:
            return child
    else:
        raise ValueError(f"child info {name} not found")


class Test(unittest.TestCase):
    def setUp(self):
        repo.require("GLib")
        repo.require("GObject")
        repo.require("GIMarshallingTests")

    def test_repo_get_dependencies(self):
        self.assertRaises(TypeError, repo.get_dependencies)
        self.assertEqual(repo.get_dependencies("GLib"), [])
        self.assertEqual(repo.get_dependencies("GObject"), ["GLib-2.0"])

    def test_repo_is_registered(self):
        self.assertRaises(TypeError, repo.is_registered)
        self.assertRaises(TypeError, repo.is_registered, None)
        self.assertTrue(repo.is_registered("GIRepository"))
        self.assertTrue(repo.is_registered("GIRepository", None))
        self.assertTrue(isinstance(repo.is_registered("GIRepository"), bool))
        self.assertTrue(repo.is_registered("GIRepository", "3.0"))
        self.assertFalse(repo.is_registered("GIRepository", ""))
        self.assertFalse(repo.is_registered("GIRepository", "99.0"))
        self.assertFalse(repo.is_registered("GIRepository", "1.0"))

    def test_repo_get_immediate_dependencies(self):
        self.assertRaises(TypeError, repo.get_immediate_dependencies)
        self.assertEqual(repo.get_immediate_dependencies("GLib"), [])
        self.assertEqual(repo.get_immediate_dependencies("GObject"), ["GLib-2.0"])
        self.assertEqual(
            repo.get_immediate_dependencies(namespace="GObject"), ["GLib-2.0"]
        )
        self.assertEqual(
            repo.get_immediate_dependencies("GIMarshallingTests"), ["Gio-2.0"]
        )

    def test_arg_info(self):
        func_info = repo.find_by_name("GIMarshallingTests", "array_fixed_out_struct")
        args = func_info.get_arguments()
        self.assertTrue(len(args), 1)

        arg = args[0]
        self.assertEqual(arg.get_container(), func_info)
        self.assertEqual(arg.get_direction(), GIRepository.Direction.OUT)
        self.assertEqual(arg.get_name(), "structs")
        self.assertEqual(arg.get_namespace(), "GIMarshallingTests")
        self.assertFalse(arg.is_caller_allocates())
        self.assertFalse(arg.is_optional())
        self.assertFalse(arg.is_return_value())
        self.assertFalse(arg.may_be_null())
        self.assertEqual(arg.get_destroy_index(), -1)
        self.assertEqual(arg.get_ownership_transfer(), GIRepository.Transfer.NOTHING)
        self.assertEqual(arg.get_scope(), GIRepository.ScopeType.INVALID)
        self.assertEqual(arg.get_type_info().get_tag(), GIRepository.TypeTag.ARRAY)

    def test_base_info(self):
        info = repo.find_by_name("GIMarshallingTests", "Object")
        self.assertEqual(info.__name__, "Object")
        self.assertEqual(info.get_name(), "Object")
        self.assertEqual(info.__module__, "gi.repository.GIMarshallingTests")
        self.assertEqual(info.get_name_unescaped(), "Object")
        self.assertEqual(info.get_namespace(), "GIMarshallingTests")
        self.assertEqual(info.get_container(), None)

        info2 = repo.find_by_name("GIMarshallingTests", "Object")
        self.assertFalse(info is info2)
        self.assertEqual(info, info2)
        self.assertTrue(info.equal(info2))
        assert isinstance(info.is_deprecated(), bool)
        assert info.get_attribute("nopenope") is None

    def test_object_info(self):
        info = repo.find_by_name("GIMarshallingTests", "Object")
        self.assertEqual(info.get_parent(), repo.find_by_name("GObject", "Object"))
        self.assertTrue(isinstance(info.get_methods(), abc.Iterable))
        self.assertTrue(isinstance(info.get_fields(), abc.Iterable))
        self.assertTrue(isinstance(info.get_interfaces(), abc.Iterable))
        self.assertTrue(isinstance(info.get_constants(), abc.Iterable))
        self.assertTrue(isinstance(info.get_vfuncs(), abc.Iterable))
        self.assertTrue(isinstance(info.get_properties(), abc.Iterable))
        self.assertFalse(info.get_abstract())
        self.assertEqual(
            info.get_class_struct(),
            repo.find_by_name("GIMarshallingTests", "ObjectClass"),
        )
        self.assertEqual(info.get_type_name(), "GIMarshallingTestsObject")
        self.assertEqual(info.get_type_init(), "gi_marshalling_tests_object_get_type")
        self.assertFalse(info.get_fundamental())
        self.assertEqual(info.get_parent(), repo.find_by_name("GObject", "Object"))
        assert info.find_vfunc("nopenope") is None
        vfunc = info.find_vfunc("method_int8_out")
        assert isinstance(vfunc, GIRepository.VFuncInfo)

    def test_callable_inheritance(self):
        self.assertTrue(issubclass(GIRepository.CallableInfo, GIRepository.BaseInfo))
        self.assertTrue(
            issubclass(GIRepository.CallbackInfo, GIRepository.CallableInfo)
        )
        self.assertTrue(
            issubclass(GIRepository.FunctionInfo, GIRepository.CallableInfo)
        )
        self.assertTrue(issubclass(GIRepository.VFuncInfo, GIRepository.CallableInfo))
        self.assertTrue(issubclass(GIRepository.SignalInfo, GIRepository.CallableInfo))

    def test_registered_type_info(self):
        info = repo.find_by_name("GIMarshallingTests", "Object")
        # Call these from the class because GIObjectInfo overrides them
        self.assertEqual(
            GIRepository.RegisteredTypeInfo.get_g_type(info),
            GObject.type_from_name("GIMarshallingTestsObject"),
        )
        self.assertEqual(
            GIRepository.RegisteredTypeInfo.get_type_name(info),
            "GIMarshallingTestsObject",
        )
        self.assertEqual(
            GIRepository.RegisteredTypeInfo.get_type_init(info),
            "gi_marshalling_tests_object_get_type",
        )

    def test_fundamental_object_info(self):
        repo.require("Regress")
        info = repo.find_by_name("Regress", "TestFundamentalObject")
        self.assertTrue(info.get_abstract())
        self.assertTrue(info.get_fundamental())
        self.assertEqual(info.get_ref_function(), "regress_test_fundamental_object_ref")
        self.assertEqual(
            info.get_unref_function(), "regress_test_fundamental_object_unref"
        )
        self.assertEqual(
            info.get_get_value_function(), "regress_test_value_get_fundamental_object"
        )
        self.assertEqual(
            info.get_set_value_function(), "regress_test_value_set_fundamental_object"
        )

    def test_interface_info(self):
        info = repo.find_by_name("GIMarshallingTests", "Interface")
        self.assertTrue(isinstance(info.get_methods(), abc.Iterable))
        self.assertTrue(isinstance(info.get_vfuncs(), abc.Iterable))
        self.assertTrue(isinstance(info.get_constants(), abc.Iterable))
        self.assertTrue(isinstance(info.get_prerequisites(), abc.Iterable))
        self.assertTrue(isinstance(info.get_properties(), abc.Iterable))
        self.assertTrue(isinstance(info.get_signals(), abc.Iterable))

        method = info.find_method("test_int8_in")
        vfunc = info.find_vfunc("test_int8_in")
        self.assertEqual(method.get_name(), "test_int8_in")
        self.assertEqual(vfunc.get_invoker(), method)
        self.assertEqual(method.get_vfunc(), vfunc)

        iface = info.get_iface_struct()
        self.assertEqual(
            iface, repo.find_by_name("GIMarshallingTests", "InterfaceIface")
        )

        assert info.find_signal("nopenope") is None

    def test_struct_info(self):
        info = repo.find_by_name("GIMarshallingTests", "InterfaceIface")
        self.assertTrue(isinstance(info, GIRepository.StructInfo))
        self.assertTrue(isinstance(info.get_fields(), abc.Iterable))
        self.assertTrue(isinstance(info.get_methods(), abc.Iterable))
        self.assertTrue(isinstance(info.get_size(), int))
        self.assertTrue(isinstance(info.get_alignment(), int))
        self.assertTrue(info.is_gtype_struct())
        self.assertFalse(info.is_foreign())

        info = repo.find_by_name("GIMarshallingTests", "SimpleStruct")
        assert info.find_method("nope") is None
        assert isinstance(info.find_method("method"), GIRepository.FunctionInfo)

        assert info.find_field("nope") is None
        assert isinstance(info.find_field("int8"), GIRepository.FieldInfo)

    def test_enum_info(self):
        info = repo.find_by_name("GIMarshallingTests", "Enum")
        self.assertTrue(isinstance(info, GIRepository.EnumInfo))
        self.assertTrue(isinstance(info.get_values(), abc.Iterable))
        self.assertTrue(isinstance(info.get_methods(), abc.Iterable))
        self.assertFalse(info.is_flags())
        self.assertTrue(info.get_storage_type() > 0)  # might be platform dependent

    def test_union_info(self):
        info = repo.find_by_name("GIMarshallingTests", "Union")
        self.assertTrue(isinstance(info, GIRepository.UnionInfo))
        self.assertTrue(isinstance(info.get_fields(), abc.Iterable))
        self.assertTrue(isinstance(info.get_methods(), abc.Iterable))
        self.assertTrue(isinstance(info.get_size(), int))
        self.assertTrue(isinstance(info.get_alignment(), int))

    def test_type_info(self):
        func_info = repo.find_by_name("GIMarshallingTests", "array_fixed_out_struct")
        (arg_info,) = func_info.get_arguments()
        type_info = arg_info.get_type_info()

        self.assertTrue(type_info.is_pointer())
        self.assertEqual(type_info.get_tag(), GIRepository.TypeTag.ARRAY)
        self.assertEqual(type_info.get_tag_as_string(), "array")
        self.assertEqual(
            type_info.get_param_type(0).get_tag(), GIRepository.TypeTag.INTERFACE
        )
        self.assertEqual(
            type_info.get_param_type(0).get_interface(),
            repo.find_by_name("GIMarshallingTests", "SimpleStruct"),
        )
        self.assertEqual(type_info.get_interface(), None)
        self.assertEqual(type_info.get_array_length_index(), -1)
        self.assertEqual(type_info.get_array_fixed_size(), 2)
        self.assertFalse(type_info.is_zero_terminated())
        self.assertEqual(type_info.get_array_type(), GIRepository.ArrayType.C)

    def test_field_info(self):
        info = repo.find_by_name("GIMarshallingTests", "InterfaceIface")
        field = find_child_info(info, "get_fields", "test_int8_in")
        self.assertEqual(field.get_name(), "test_int8_in")
        self.assertTrue(field.get_flags() & GIRepository.FieldInfoFlags.IS_READABLE)
        self.assertFalse(field.get_flags() & GIRepository.FieldInfoFlags.IS_WRITABLE)
        self.assertEqual(
            field.get_type_info().get_tag(), GIRepository.TypeTag.INTERFACE
        )

        # don't test actual values because that might fail with architecture differences
        self.assertTrue(isinstance(field.get_size(), int))
        self.assertTrue(isinstance(field.get_offset(), int))

    def test_property_info(self):
        info = repo.find_by_name("GIMarshallingTests", "PropertiesObject")
        prop = find_child_info(info, "get_properties", "some-object")

        flags = (
            GObject.ParamFlags.READABLE
            | GObject.ParamFlags.WRITABLE
            | GObject.ParamFlags.CONSTRUCT
        )
        self.assertEqual(prop.get_flags(), flags)
        self.assertEqual(prop.get_type_info().get_tag(), GIRepository.TypeTag.INTERFACE)
        self.assertEqual(
            prop.get_type_info().get_interface(), repo.find_by_name("GObject", "Object")
        )
        self.assertEqual(prop.get_ownership_transfer(), GIRepository.Transfer.NOTHING)

    def test_callable_info(self):
        func_info = repo.find_by_name("GIMarshallingTests", "array_fixed_out_struct")
        self.assertTrue(hasattr(func_info, "invoke"))
        self.assertTrue(isinstance(func_info.get_arguments(), abc.Iterable))
        self.assertEqual(func_info.get_caller_owns(), GIRepository.Transfer.NOTHING)
        self.assertFalse(func_info.may_return_null())
        self.assertEqual(
            func_info.get_return_type().get_tag(), GIRepository.TypeTag.VOID
        )
        self.assertRaises(
            AttributeError, func_info.get_return_attribute, "_not_an_attr"
        )

    def test_signal_info(self):
        repo.require("Regress")
        info = repo.find_by_name("Regress", "TestObj")
        sig_info = find_child_info(info, "get_signals", "test")

        sig_flags = (
            GObject.SignalFlags.RUN_LAST
            | GObject.SignalFlags.NO_RECURSE
            | GObject.SignalFlags.NO_HOOKS
        )

        self.assertTrue(sig_info is not None)
        self.assertTrue(isinstance(sig_info, GIRepository.CallableInfo))
        self.assertTrue(isinstance(sig_info, GIRepository.SignalInfo))
        self.assertEqual(sig_info.get_name(), "test")
        self.assertEqual(sig_info.get_class_closure(), None)
        self.assertFalse(sig_info.true_stops_emit())
        self.assertEqual(sig_info.get_flags(), sig_flags)

    def test_notify_signal_info_with_obj(self):
        repo.require("Regress")
        info = repo.find_by_name("Regress", "TestObj")
        sig_info = find_child_info(info, "get_signals", "sig-with-array-prop")

        sig_flags = GObject.SignalFlags.RUN_LAST

        self.assertTrue(sig_info is not None)
        self.assertTrue(isinstance(sig_info, GIRepository.CallableInfo))
        self.assertTrue(isinstance(sig_info, GIRepository.SignalInfo))
        self.assertEqual(sig_info.get_name(), "sig-with-array-prop")
        self.assertEqual(sig_info.get_class_closure(), None)
        self.assertFalse(sig_info.true_stops_emit())
        self.assertEqual(sig_info.get_flags(), sig_flags)

    def test_object_constructor(self):
        info = repo.find_by_name("GIMarshallingTests", "Object")
        method = find_child_info(info, "get_methods", "new")

        self.assertTrue(isinstance(method, GIRepository.CallableInfo))
        self.assertTrue(isinstance(method, GIRepository.FunctionInfo))
        self.assertTrue(method in info.get_methods())
        self.assertEqual(method.get_name(), "new")
        self.assertFalse(method.is_method())
        self.assertTrue(method.is_constructor())
        self.assertEqual(method.get_symbol(), "gi_marshalling_tests_object_new")

        flags = method.get_flags()
        self.assertFalse(flags & GIRepository.FunctionInfoFlags.IS_METHOD)
        self.assertTrue(flags & GIRepository.FunctionInfoFlags.IS_CONSTRUCTOR)
        self.assertFalse(flags & GIRepository.FunctionInfoFlags.IS_GETTER)
        self.assertFalse(flags & GIRepository.FunctionInfoFlags.IS_SETTER)
        self.assertFalse(flags & GIRepository.FunctionInfoFlags.WRAPS_VFUNC)

    def test_method_info(self):
        info = repo.find_by_name("GIMarshallingTests", "Object")
        method = find_child_info(info, "get_methods", "vfunc_return_value_only")

        self.assertTrue(isinstance(method, GIRepository.CallableInfo))
        self.assertTrue(isinstance(method, GIRepository.FunctionInfo))
        self.assertTrue(method in info.get_methods())
        self.assertEqual(method.get_name(), "vfunc_return_value_only")
        self.assertFalse(method.is_constructor())
        self.assertEqual(
            method.get_symbol(), "gi_marshalling_tests_object_vfunc_return_value_only"
        )
        self.assertTrue(method.is_method())

        flags = method.get_flags()
        self.assertTrue(flags & GIRepository.FunctionInfoFlags.IS_METHOD)
        self.assertFalse(flags & GIRepository.FunctionInfoFlags.IS_CONSTRUCTOR)
        self.assertFalse(flags & GIRepository.FunctionInfoFlags.IS_GETTER)
        self.assertFalse(flags & GIRepository.FunctionInfoFlags.IS_SETTER)
        self.assertFalse(flags & GIRepository.FunctionInfoFlags.WRAPS_VFUNC)

    def test_method_finish_func(self):
        repo.require("Regress")
        info = repo.find_by_name("Regress", "TestObj")
        method = find_child_info(info, "get_methods", "function_finish")
        assert method.get_finish_func() is None

    def test_async_method_finish_func(self):
        repo.require("Regress")
        info = repo.find_by_name("Regress", "TestObj")
        method = find_child_info(info, "get_methods", "function_async")
        finish_func = method.get_finish_func()
        assert finish_func.get_name() == "function_finish"

    def test_vfunc_info(self):
        info = repo.find_by_name("GIMarshallingTests", "Object")
        invoker = find_child_info(info, "get_methods", "vfunc_return_value_only")
        vfunc = find_child_info(info, "get_vfuncs", "vfunc_return_value_only")

        self.assertTrue(isinstance(vfunc, GIRepository.CallableInfo))
        self.assertTrue(isinstance(vfunc, GIRepository.VFuncInfo))
        self.assertEqual(vfunc.get_name(), "vfunc_return_value_only")
        self.assertEqual(vfunc.get_invoker(), invoker)
        self.assertEqual(invoker, info.find_method("vfunc_return_value_only"))
        self.assertEqual(vfunc.get_flags(), 0)
        self.assertEqual(vfunc.get_offset(), 0xFFFF)  # unknown offset
        self.assertEqual(vfunc.get_signal(), None)

    def test_callable_can_throw_gerror(self):
        info = repo.find_by_name("GIMarshallingTests", "Object")
        invoker = find_child_info(info, "get_methods", "vfunc_meth_with_error")
        vfunc = find_child_info(info, "get_vfuncs", "vfunc_meth_with_err")

        self.assertTrue(invoker.can_throw_gerror())
        self.assertTrue(vfunc.can_throw_gerror())

        # Test that args do not include the GError**
        self.assertEqual(len(invoker.get_arguments()), 1)
        self.assertEqual(len(vfunc.get_arguments()), 1)

        # Sanity check method that does not throw.
        invoker_no_throws = find_child_info(info, "get_methods", "method_int8_in")
        self.assertFalse(invoker_no_throws.can_throw_gerror())

    def test_enums(self):
        self.assertTrue(hasattr(GIRepository, "Direction"))
        self.assertTrue(hasattr(GIRepository, "Transfer"))
        self.assertTrue(hasattr(GIRepository, "ArrayType"))
        self.assertTrue(hasattr(GIRepository, "ScopeType"))
        self.assertTrue(hasattr(GIRepository, "VFuncInfoFlags"))
        self.assertTrue(hasattr(GIRepository, "FieldInfoFlags"))
        self.assertTrue(hasattr(GIRepository, "FunctionInfoFlags"))
        self.assertTrue(hasattr(GIRepository, "TypeTag"))

    def test_introspected_argument_info(self):
        self.assertTrue(
            isinstance(IntrospectedRepository.Argument.__info__, GIRepository.UnionInfo)
        )

        arg = IntrospectedRepository.Argument()
        self.assertTrue(isinstance(arg.__info__, GIRepository.UnionInfo))

        old_info = IntrospectedRepository.Argument.__info__
        IntrospectedRepository.Argument.__info__ = "not an info"
        self.assertRaises(TypeError, IntrospectedRepository.Argument)
        IntrospectedRepository.Argument.__info__ = old_info



# Copyright (C) 2012 Martin Pitt <martinpitt@gnome.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

from ..importer import get_introspection_module
from ..overrides import override

Regress = get_introspection_module("Regress")

REGRESS_OVERRIDE = 42


class Bitmask(Regress.Bitmask):
    """Replicate override of Bitmask (uint64), similar to GStreamer."""

    def __init__(self, v):
        if not isinstance(v, int):
            raise TypeError(f"{type(v)} is not an int.")

        self.v = int(v)

    def __str__(self):
        return hex(self.v)

    def __eq__(self, other):
        return self.v == other


Bitmask = override(Bitmask)

__all__ = ["REGRESS_OVERRIDE", "Bitmask"]



import gc
import weakref

import pytest
from gi.repository import GObject, Regress

try:
    from gi.repository import Gtk

    GTK4 = Gtk._version == "4.0"
except ImportError:
    Gtk = None
    GTK4 = False


def test_constructor_no_data():
    obj = Regress.TestFundamentalSubObject()

    assert isinstance(obj, Regress.TestFundamentalSubObject)
    assert isinstance(obj, Regress.TestFundamentalObject)
    assert obj.refcount == 1
    assert obj.data is None


def test_constructor_with_data():
    with pytest.raises(TypeError):
        Regress.TestFundamentalSubObject(data="foo")


def test_create_fundamental_new_with_data():
    obj = Regress.TestFundamentalSubObject.new("foo")

    assert isinstance(obj, Regress.TestFundamentalSubObject)
    assert isinstance(obj, Regress.TestFundamentalObject)
    assert obj.refcount == 1
    assert obj.data == "foo"


def test_change_field():
    obj = Regress.TestFundamentalObjectNoGetSetFunc.new("foo")

    obj.data = "bar"

    assert obj.get_data() == "bar"


def test_call_method():
    obj = Regress.TestFundamentalObjectNoGetSetFunc.new("foo")

    assert obj.get_data() == "foo"


def test_create_fundamental_hidden_class_instance():
    obj = Regress.test_create_fundamental_hidden_class_instance()

    assert isinstance(obj, Regress.TestFundamentalObject)


def test_create_fundamental_refcount():
    obj = Regress.TestFundamentalSubObject.new("foo")

    assert obj.refcount == 1


def test_delete_fundamental_refcount():
    obj = Regress.TestFundamentalSubObject.new("foo")
    del obj

    gc.collect()


def test_value_set_fundamental_object():
    val = GObject.Value(Regress.TestFundamentalSubObject.__gtype__)
    obj = Regress.TestFundamentalSubObject()

    val.set_value(obj)

    assert val.get_value() == obj


def test_value_set_wrong_value():
    val = GObject.Value(Regress.TestFundamentalSubObject.__gtype__)

    with pytest.raises(TypeError, match="Fundamental type is required"):
        val.set_value(1)


def test_value_set_wrong_fundamental():
    class MyCustomFundamentalObject(Regress.TestFundamentalObject):
        pass

    val = GObject.Value(Regress.TestFundamentalSubObject.__gtype__)

    with pytest.raises(TypeError, match="Invalid fundamental type for assignment"):
        val.set_value(MyCustomFundamentalObject())


def test_array_of_fundamental_objects_in():
    assert Regress.test_array_of_fundamental_objects_in(
        [Regress.TestFundamentalSubObject()]
    )


def test_array_of_fundamental_objects_out():
    objs = Regress.test_array_of_fundamental_objects_out()

    assert len(objs) == 2
    assert all(isinstance(o, Regress.TestFundamentalObject) for o in objs)


def test_fundamental_argument_in():
    obj = Regress.TestFundamentalSubObject()

    assert Regress.test_fundamental_argument_in(obj)


def test_abstract_fundamental_type():
    with pytest.raises(TypeError):
        Regress.TestFundamentalObject()


def test_fundamental_argument_out():
    obj = Regress.TestFundamentalSubObject.new("data")
    other = Regress.test_fundamental_argument_out(obj)

    assert type(obj) is type(other)
    assert obj is not other
    assert obj.data == other.data


def test_multiple_objects():
    obj1 = Regress.TestFundamentalSubObject()
    obj2 = Regress.TestFundamentalSubObject()

    assert obj1 != obj2


def test_fundamental_weak_ref():
    obj = Regress.TestFundamentalSubObject()
    weak = weakref.ref(obj)

    assert weak() == obj

    del obj
    gc.collect()

    assert weak() is None


def test_fundamental_primitive_object():
    bitmask = Regress.Bitmask(2)

    assert bitmask.v == 2


def test_custom_fundamental_type_vfunc_override():
    class MyCustomFundamentalObject(Regress.TestFundamentalObject):
        def __init__(self):
            super().__init__()
            invocations["__init__"] = True

        def do_finalize(self):
            invocations["do_finalize"] = True

    invocations = {}
    obj = MyCustomFundamentalObject()
    del obj
    gc.collect()

    assert "__init__" in invocations
    assert "do_finalize" in invocations


@pytest.mark.skipif(not GTK4, reason="requires GTK 4")
def test_gtk_expression():
    obj = object()
    con = Gtk.ConstantExpression.new_for_value(obj)

    assert con.get_value() is obj


@pytest.mark.skipif(not GTK4, reason="requires GTK 4")
def test_gtk_string_filter_fundamental_property():
    expr = Gtk.ConstantExpression.new_for_value("one")
    filter = Gtk.StringFilter.new(expr)
    filter.props.expression = expr

    assert filter.get_expression() == expr
    assert filter.props.expression == expr



import tempfile
import os
import weakref

import pytest

Gtk = pytest.importorskip("gi.repository.Gtk")
GLib = pytest.importorskip("gi.repository.GLib")
GObject = pytest.importorskip("gi.repository.GObject")
Gio = pytest.importorskip("gi.repository.Gio")

from .helper import capture_exceptions

GTK4 = Gtk._version == "4.0"


def new_gtype_name(_count=[0]):
    _count[0] += 1
    return f"GtkTemplateTest{_count[0]:d}"


def ensure_resource_registered():
    resource_path = "/org/gnome/pygobject/test/a.ui"

    def is_registered(path):
        try:
            Gio.resources_get_info(path, Gio.ResourceLookupFlags.NONE)
        except GLib.Error:
            return False
        return True

    if is_registered(resource_path):
        return resource_path

    gresource_data = (
        b"GVariant\x00\x00\x00\x00\x00\x00\x00\x00\x18\x00\x00\x00"
        b"\xc8\x00\x00\x00\x00\x00\x00(\x06\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x01\x00\x00\x00\x04\x00\x00\x00\x05\x00\x00\x00"
        b"\x06\x00\x00\x00KP\x90\x0b\x03\x00\x00\x00\xc8\x00\x00\x00"
        b"\x04\x00L\x00\xcc\x00\x00\x00\xd0\x00\x00\x00\xb0\xb7$0"
        b"\x00\x00\x00\x00\xd0\x00\x00\x00\x06\x00L\x00\xd8\x00\x00\x00"
        b"\xdc\x00\x00\x00f\xc30\xd1\x01\x00\x00\x00\xdc\x00\x00\x00"
        b"\n\x00L\x00\xe8\x00\x00\x00\xec\x00\x00\x00\xd4\xb5\x02\x00"
        b"\xff\xff\xff\xff\xec\x00\x00\x00\x01\x00L\x00\xf0\x00\x00\x00"
        b"\xf4\x00\x00\x005H}\xe3\x02\x00\x00\x00\xf4\x00\x00\x00"
        b"\x05\x00L\x00\xfc\x00\x00\x00\x00\x01\x00\x00\xa2^\xd6t"
        b"\x04\x00\x00\x00\x00\x01\x00\x00\x04\x00v\x00\x08\x01\x00\x00"
        b"\xa5\x01\x00\x00org/\x01\x00\x00\x00gnome/\x00\x00\x02\x00\x00\x00"
        b"pygobject/\x00\x00\x04\x00\x00\x00/\x00\x00\x00\x00\x00\x00\x00"
        b"test/\x00\x00\x00\x05\x00\x00\x00a.ui\x00\x00\x00\x00"
        b'\x8d\x00\x00\x00\x00\x00\x00\x00<interface>\n  <template class="G'
        b'tkTemplateTestResource" parent="GtkBox">\n  <property name="spaci'
        b'ng">42</property>\n  </template>\n</interface>\n\x00\x00(uuay)'
    )

    resource = Gio.Resource.new_from_data(GLib.Bytes.new(gresource_data))
    Gio.resources_register(resource)
    assert is_registered(resource_path)
    return resource_path


def test_allow_init_template_call():
    type_name = new_gtype_name()

    xml = f"""\
<interface>
  <template class="{type_name}" parent="GtkBox">
  </template>
</interface>
"""

    @Gtk.Template.from_string(xml)
    class Foo(Gtk.Box):
        __gtype_name__ = type_name

        def __init__(self):
            super().__init__()
            self.init_template()

    # Stop current pygobject from handling the initialisation
    del Foo.__dontuse_ginstance_init__

    Foo()


def test_init_template_second_instance():
    type_name = new_gtype_name()

    xml = f"""\
<interface>
  <template class="{type_name}" parent="GtkBox">
    <child>
      <object class="GtkLabel" id="label">
      </object>
    </child>
  </template>
</interface>
"""

    @Gtk.Template.from_string(xml)
    class Foo(Gtk.Box):
        __gtype_name__ = type_name

        label = Gtk.Template.Child("label")

        def __init__(self):
            super().__init__()
            self.init_template()

    # Stop current pygobject from handling the initialisation
    del Foo.__dontuse_ginstance_init__

    foo = Foo()
    assert isinstance(foo.label, Gtk.Label)

    foo2 = Foo()
    assert isinstance(foo2.label, Gtk.Label)


def test_main_example():
    type_name = new_gtype_name()

    example_xml = f"""\
<interface>
  <template class="{type_name}" parent="GtkBox">
    <property name="orientation">GTK_ORIENTATION_HORIZONTAL</property>
    <property name="spacing">4</property>
    <child>
      <object class="GtkButton" id="hello_button">
        <property name="label">Hello World</property>
        <signal name="clicked" handler="hello_button_clicked"
                object="{type_name}" swapped="no"/>
        <signal name="clicked" handler="hello_button_clicked_after"
                object="{type_name}" swapped="no" after="yes"/>
      </object>
    </child>
    <child>
      <object class="GtkButton" id="goodbye_button">
        <property name="label">Goodbye World</property>
        <signal name="clicked" handler="goodbye_button_clicked"/>
        <signal name="clicked" handler="goodbye_button_clicked_after"
                after="yes"/>
      </object>
    </child>
  </template>
</interface>
"""

    @Gtk.Template.from_string(example_xml)
    class Foo(Gtk.Box):
        __gtype_name__ = type_name

        def __init__(self):
            super().__init__()
            self.callback_hello = []
            self.callback_hello_after = []
            self.callback_goodbye = []
            self.callback_goodbye_after = []

        @Gtk.Template.Callback("hello_button_clicked")
        def _hello_button_clicked(self, *args):
            self.callback_hello.append(args)

        @Gtk.Template.Callback("hello_button_clicked_after")
        def _hello_after(self, *args):
            self.callback_hello_after.append(args)

        _hello_button = Gtk.Template.Child("hello_button")

        goodbye_button = Gtk.Template.Child()

        @Gtk.Template.Callback("goodbye_button_clicked")
        def _goodbye_button_clicked(self, *args):
            self.callback_goodbye.append(args)

        @Gtk.Template.Callback("goodbye_button_clicked_after")
        def _goodbye_after(self, *args):
            self.callback_goodbye_after.append(args)

    w = Foo()
    assert w.__gtype__.name == type_name
    assert w.props.orientation == Gtk.Orientation.HORIZONTAL
    assert w.props.spacing == 4
    assert isinstance(w._hello_button, Gtk.Button)
    assert w._hello_button.props.label == "Hello World"
    assert isinstance(w.goodbye_button, Gtk.Button)
    assert w.goodbye_button.props.label == "Goodbye World"

    assert w.callback_hello == []
    w._hello_button.emit("clicked")
    assert w.callback_hello == [(w,)]
    assert w.callback_hello_after == [(w,)]

    assert w.callback_goodbye == []
    w.goodbye_button.emit("clicked")
    assert w.callback_goodbye == [(w.goodbye_button,)]
    assert w.callback_goodbye_after == [(w.goodbye_button,)]


def test_duplicate_handler():
    type_name = new_gtype_name()

    xml = f"""\
<interface>
  <template class="{type_name}" parent="GtkBox">
    <child>
      <object class="GtkButton" id="hello_button">
        <signal name="clicked" handler="hello_button_clicked" />
      </object>
    </child>
  </template>
</interface>
"""

    class Foo(Gtk.Box):
        __gtype_name__ = type_name

        @Gtk.Template.Callback("hello_button_clicked")
        def _hello_button_clicked(self, *args):
            pass

        @Gtk.Template.Callback()
        def hello_button_clicked(self, *args):
            pass

    with pytest.raises(RuntimeError, match=r".*hello_button_clicked.*"):
        Gtk.Template.from_string(xml)(Foo)


def test_duplicate_child():
    type_name = new_gtype_name()

    xml = f"""\
<interface>
  <template class="{type_name}" parent="GtkBox">
    <child>
      <object class="GtkButton" id="hello_button" />
    </child>
  </template>
</interface>
"""

    class Foo(Gtk.Box):
        __gtype_name__ = type_name

        foo = Gtk.Template.Child("hello_button")
        hello_button = Gtk.Template.Child()

    with pytest.raises(RuntimeError, match=r".*hello_button.*"):
        Gtk.Template.from_string(xml)(Foo)


def test_nonexist_handler():
    type_name = new_gtype_name()

    xml = f"""\
<interface>
  <template class="{type_name}" parent="GtkBox">
  </template>
</interface>
"""

    @Gtk.Template.from_string(xml)
    class Foo(Gtk.Box):
        __gtype_name__ = type_name

        @Gtk.Template.Callback("nonexit")
        def foo(self, *args):
            pass

    with capture_exceptions() as exc_info:
        Foo()
    assert "nonexit" in str(exc_info[0].value)
    assert exc_info[0].type is RuntimeError


def test_missing_handler_callback():
    type_name = new_gtype_name()

    xml = f"""\
<interface>
  <template class="{type_name}" parent="GtkBox">
    <child>
      <object class="GtkButton" id="hello_button">
        <signal name="clicked" handler="i_am_not_used_in_python" />
      </object>
    </child>
  </template>
</interface>
"""

    class Foo(Gtk.Box):
        __gtype_name__ = type_name

    Gtk.Template.from_string(xml)(Foo)()


def test_handler_swapped_not_supported():
    type_name = new_gtype_name()

    xml = f"""\
<interface>
  <template class="{type_name}" parent="GtkBox">
    <child>
      <object class="GtkButton" id="hello_button">
        <signal name="clicked" handler="hello_button_clicked"
                object="{type_name}" swapped="yes" />
      </object>
    </child>
  </template>
</interface>
"""

    @Gtk.Template.from_string(xml)
    class Foo(Gtk.Box):
        __gtype_name__ = type_name

        hello_button = Gtk.Template.Child()

        @Gtk.Template.Callback("hello_button_clicked")
        def foo(self, *args):
            pass

    with capture_exceptions() as exc_info:
        Foo()
    assert "ConnectFlags.SWAPPED" in str(exc_info[0].value)


def test_handler_class_staticmethod():
    type_name = new_gtype_name()

    xml = f"""\
<interface>
  <template class="{type_name}" parent="GtkBox">
    <child>
      <object class="GtkButton" id="hello_button">
        <signal name="clicked" handler="clicked_class" />
        <signal name="clicked" handler="clicked_static" />
      </object>
    </child>
  </template>
</interface>
"""

    signal_args_class = []
    signal_args_static = []

    @Gtk.Template.from_string(xml)
    class Foo(Gtk.Box):
        __gtype_name__ = type_name

        hello_button = Gtk.Template.Child()

        @Gtk.Template.Callback("clicked_class")
        @classmethod
        def cb1(*args):
            signal_args_class.append(args)

        @Gtk.Template.Callback("clicked_static")
        @staticmethod
        def cb2(*args):
            signal_args_static.append(args)

    foo = Foo()
    foo.hello_button.emit("clicked")
    assert signal_args_class == [(Foo, foo.hello_button)]
    assert signal_args_static == [(foo.hello_button,)]


@pytest.mark.skipif(Gtk._version == "4.0", reason="errors out first with gtk4")
def test_check_decorated_class():
    NonWidget = type("Foo", (object,), {})
    with pytest.raises(TypeError, match=r".*on Widgets.*"):
        Gtk.Template.from_string("")(NonWidget)

    Widget = type("Foo", (Gtk.Widget,), {"__gtype_name__": new_gtype_name()})
    with pytest.raises(TypeError, match=r".*Cannot nest.*"):
        Gtk.Template.from_string("")(Gtk.Template.from_string("")(Widget))

    Widget = type("Foo", (Gtk.Widget,), {})
    with pytest.raises(TypeError, match=r".*__gtype_name__.*"):
        Gtk.Template.from_string("")(Widget)

    with pytest.raises(TypeError, match=r".*on Widgets.*"):
        Gtk.Template.from_string("")(object())


@pytest.mark.skipif(Gtk._version == "4.0", reason="errors out first with gtk4")
def test_subclass_fail():
    @Gtk.Template.from_string("")
    class Base(Gtk.Widget):
        __gtype_name__ = new_gtype_name()

    with capture_exceptions() as exc_info:
        type("Sub", (Base,), {})()
    assert "not allowed at this time" in str(exc_info[0].value)
    assert exc_info[0].type is TypeError


def test_from_file():
    fd, name = tempfile.mkstemp()
    try:
        os.close(fd)

        type_name = new_gtype_name()

        with open(name, "wb") as h:
            h.write(
                f"""\
    <interface>
      <template class="{type_name}" parent="GtkBox">
      <property name="spacing">42</property>
      </template>
    </interface>
    """.encode()
            )

        @Gtk.Template.from_file(name)
        class Foo(Gtk.Box):
            __gtype_name__ = type_name

        foo = Foo()
        assert foo.props.spacing == 42
    finally:
        os.remove(name)


def test_property_override():
    type_name = new_gtype_name()

    xml = f"""\
    <interface>
      <template class="{type_name}" parent="GtkBox">
      <property name="spacing">42</property>
      </template>
    </interface>
"""

    @Gtk.Template.from_string(xml)
    class Foo(Gtk.Box):
        __gtype_name__ = type_name

    foo = Foo()
    assert foo.props.spacing == 42

    foo = Foo(spacing=124)
    assert foo.props.spacing == 124


def test_from_file_non_exist():
    dirname = tempfile.mkdtemp()
    try:
        path = os.path.join(dirname, "noexist")

        Widget = type("Foo", (Gtk.Widget,), {"__gtype_name__": new_gtype_name()})
        with pytest.raises(GLib.Error, match=r".*No such file.*"):
            Gtk.Template.from_file(path)(Widget)
    finally:
        os.rmdir(dirname)


def test_from_string_bytes():
    type_name = new_gtype_name()

    xml = f"""\
    <interface>
      <template class="{type_name}" parent="GtkBox">
      <property name="spacing">42</property>
      </template>
    </interface>
    """.encode()

    @Gtk.Template.from_string(xml)
    class Foo(Gtk.Box):
        __gtype_name__ = type_name

    foo = Foo()
    assert foo.props.spacing == 42


def test_from_resource():
    resource_path = ensure_resource_registered()

    @Gtk.Template.from_resource(resource_path)
    class Foo(Gtk.Box):
        __gtype_name__ = "GtkTemplateTestResource"

    foo = Foo()
    assert foo.props.spacing == 42


def test_from_resource_non_exit():
    Widget = type("Foo", (Gtk.Widget,), {"__gtype_name__": new_gtype_name()})
    with pytest.raises(GLib.Error, match=r".*/or/gnome/pygobject/noexit.*"):
        Gtk.Template.from_resource("/or/gnome/pygobject/noexit")(Widget)


def test_constructors():
    with pytest.raises(TypeError):
        Gtk.Template()

    with pytest.raises(TypeError):
        Gtk.Template(foo=1)

    Gtk.Template(filename="foo")
    Gtk.Template(resource_path="foo")
    Gtk.Template(string="foo")

    with pytest.raises(TypeError):
        Gtk.Template(filename="foo", resource_path="bar")

    with pytest.raises(TypeError):
        Gtk.Template(filename="foo", nope="bar")

    Gtk.Template.from_string("bla")
    Gtk.Template.from_resource("foo")
    Gtk.Template.from_file("foo")


def test_child_construct():
    Gtk.Template.Child()
    Gtk.Template.Child("name")
    with pytest.raises(TypeError):
        Gtk.Template.Child("name", True)
    Gtk.Template.Child("name", internal=True)
    with pytest.raises(TypeError):
        Gtk.Template.Child("name", internal=True, something=False)


def test_internal_child():
    main_type_name = new_gtype_name()

    xml = f"""\
    <interface>
      <template class="{main_type_name}" parent="GtkBox">
        <child>
          <object class="GtkBox" id="somechild">
            <property name="margin-top">42</property>
          </object>
        </child>
      </template>
    </interface>
    """

    @Gtk.Template.from_string(xml)
    class MainThing(Gtk.Box):
        __gtype_name__ = main_type_name

        somechild = Gtk.Template.Child(internal=True)

    thing = MainThing()
    assert thing.somechild.props.margin_top == 42

    other_type_name = new_gtype_name()

    xml = f"""\
    <interface>
      <template class="{other_type_name}" parent="GtkBox">
        <child>
          <object class="{main_type_name}">
            <child internal-child="somechild">
              <object class="GtkBox">
                <property name="margin-top">24</property>
                <child>
                  <object class="GtkLabel">
                    <property name="label">foo</property>
                  </object>
                </child>
              </object>
            </child>
          </object>
        </child>
      </template>
    </interface>
    """

    @Gtk.Template.from_string(xml)
    class OtherThing(Gtk.Box):
        __gtype_name__ = other_type_name

    other = OtherThing()
    child = other.get_children()[0] if not GTK4 else other.get_first_child()

    assert isinstance(child, MainThing)

    child = child.get_children()[0] if not GTK4 else child.get_first_child()

    assert isinstance(child, Gtk.Box)
    assert child.props.margin_top == 24

    child = child.get_children()[0] if not GTK4 else child.get_first_child()

    assert isinstance(child, Gtk.Label)
    assert child.props.label == "foo"


def test_template_hierarchy():
    testlabel = """
    <interface>
      <template class="TestLabel" parent="GtkLabel">
      </template>
     </interface>
    """

    @Gtk.Template(string=testlabel)
    class TestLabel(Gtk.Label):
        __gtype_name__ = "TestLabel"

        def __init__(self):
            super().__init__()
            self.props.label = "TestLabel"

    testbox = """
    <interface>
      <template class="TestBox" parent="GtkBox">
        <child>
          <object class="TestLabel" id="_testlabel"/>
        </child>
      </template>
    </interface>
    """

    @Gtk.Template(string=testbox)
    class TestBox(Gtk.Box):
        __gtype_name__ = "TestBox"

        _testlabel = Gtk.Template.Child()

        def __init__(self):
            super().__init__()

            assert isinstance(self._testlabel, TestLabel)

    window = """
    <interface>
      <template class="MyWindow" parent="GtkWindow">
        <property name="title">"Hellow World"</property>
        <child>
          <object class="TestBox" id="_testbox">
            <child>
              <object class="TestLabel" id="_testlabel"/>
            </child>
          </object>
        </child>
      </template>
    </interface>
    """

    @Gtk.Template(string=window)
    class MyWindow(Gtk.Window):
        __gtype_name__ = "MyWindow"

        _testbox = Gtk.Template.Child()
        _testlabel = Gtk.Template.Child()

        def __init__(self):
            super().__init__()

            assert isinstance(self._testbox, TestBox)
            assert isinstance(self._testlabel, TestLabel)
            children = self._testbox.get_children() if not GTK4 else list(self._testbox)

            assert len(children) == 2

    win = MyWindow()
    assert isinstance(win, MyWindow)


def test_multiple_init_template_calls():
    xml = """
    <interface>
      <template class="MyBox" parent="GtkBox">
        <child>
          <object class="GtkLabel" id="_label"/>
        </child>
       </template>
     </interface>
    """

    @Gtk.Template(string=xml)
    class MyBox(Gtk.Box):
        __gtype_name__ = "MyBox"

        _label = Gtk.Template.Child()

        def __init__(self):
            super().__init__()
            self._label.props.label = "awesome label"

    my_box = MyBox()
    assert isinstance(my_box, MyBox)
    children = my_box.get_children() if not GTK4 else list(my_box)

    assert len(children) == 1
    my_box.init_template()
    assert isinstance(my_box, MyBox)
    children = my_box.get_children() if not GTK4 else list(my_box)

    assert len(children) == 1


def test_python_class_hierarchy():
    xml = """
    <interface>
      <template class="ChildBox" parent="ParentBox">
        <child>
          <object class="GtkLabel" id="_label"/>
        </child>
      </template>
    </interface>
    """

    class ParentBox(Gtk.Box):
        __gtype_name__ = "ParentBox"

        def __init__(self):
            super().__init__()

    @Gtk.Template(string=xml)
    class ChildBox(ParentBox):
        __gtype_name__ = "ChildBox"

        _label = Gtk.Template.Child()

        def __init__(self):
            super().__init__()

    childBox = ChildBox()
    assert isinstance(childBox, ChildBox)
    children = childBox.get_children() if not GTK4 else list(childBox)

    assert len(children) == 1


@pytest.mark.skipif(not GTK4, reason="GTK 4 only")
def test_finalization_of_custom_child_objects():
    xml = """
    <?xml version="1.0" encoding="UTF-8"?>
    <interface>
    <requires lib="gtk" version="4.0" />
    <template class="CustomBox" parent="GtkBox">
        <child>
            <object class="CustomLabel">
                <property name="label">CustomLabel</property>
            </object>
        </child>
    </template>
    </interface>
    """

    class CustomLabel(Gtk.Label):
        __gtype_name__ = "CustomLabel"

        def do_unroot(self) -> None:
            check_finalize(self)

    @Gtk.Template(string=xml)
    class CustomBox(Gtk.Box):
        __gtype_name__ = "CustomBox"

        def do_unroot(self) -> None:
            check_finalize(self)

    finalized = []

    def check_finalize(obj) -> None:
        name = obj.__class__.__name__
        obj.weak_ref(lambda: finalized.append(f"{name} -> GObject finalized"))
        weakref.finalize(obj, finalized.append, f"{name} -> PyObject finalized")

    custom_box = CustomBox()
    box = Gtk.Box()
    box.append(custom_box)
    window = Gtk.Window()
    window.set_child(box)
    window.present()
    box.remove(custom_box)

    del custom_box

    assert "CustomBox -> PyObject finalized" in finalized
    assert "CustomBox -> GObject finalized" in finalized
    assert "CustomLabel -> PyObject finalized" in finalized
    assert "CustomLabel -> GObject finalized" in finalized
